// -*- mode: objc; tab-width: 4 -*-

#import  "HistoryText.h"
#import  "textActions.h"
#import  "KAutomataForX.h"

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiTextP.h>
#include <X11/Xaw/TextSink.h>

static void HTextEventHandler(Widget w, HistoryText *self, XEvent *ev,
							  Boolean *continue_to_dispatch);

@implementation HistoryText

- initFromKDF:(char *)kdf_path withWidget:(Widget)aText translator:(Widget)tw
{
	if ((self = [super initFromKDF:kdf_path withWidget:aText translator:tw])) {
		XtRemoveEventHandler(text, text_event_mask, False,
							 (XtEventHandler)_KTextEventHandler, self);
		// Make text `transparent' to Key events so that Key events propagate
		// through text to arrive at focus_window in XIMPreeditPosition style.
		text_event_mask &= ~(KeyPressMask | KeyReleaseMask);
		XtAddEventHandler(text, text_event_mask, False,
						  (XtEventHandler)HTextEventHandler, self);
		sink = ((AsciiWidget)text)->text.sink;
		XtVaGetValues(text,
					  XtNleftMargin, &left_margin,
					  XtNrightMargin, &right_margin,
					  XtNwidth, &max_width, NULL);
		return self;
	}

	[self free];
	return nil;
}

- receiveEvent:(XEvent *)ev
{
	_eReceiver = self;
	XtDispatchEventToWidget(translator, ev);

	if (ev->type == KeyPress) {
		if (resize)
			[self resize];
		[self makeCaretVisible];
	}

	return self;
}

- (char *)getComposingChar
{
	return [am getComposingChar];
}

- (EventMask)getTextEventMask
{
	return text_event_mask;
}

- setLeftMargin:(Position)num
{
	left_margin = num;
	return [self makeCaretVisible];
}

- setRightMargin:(Position)num
{
	right_margin = num;
	return [self makeCaretVisible];
}

- (KeyboardType)getKeyboard
{
	return keyboard;
}

- setString:(char *)string
{
	return [[super setString:string] makeCaretVisible];
}

- setMaxWidth:(Dimension)width
{
	Dimension current_width;

	max_width = width;
	XtVaGetValues(text, XtNwidth, &current_width, NULL);
	if (current_width > max_width) {
		XtVaSetValues(text, XtNwidth, max_width, NULL);
		[self makeCaretVisible];
	}
	return self;
}

- resize
{
	Dimension width;
	Dimension text_width;
	Dimension good_width;
	Position widget_left_margin;
	XFontSet font_set;
	char *string;
	
	XtVaGetValues(text, XtNleftMargin, &widget_left_margin,
				  XtNfontSet, &font_set, XtNwidth, &width,
				  XtNstring, &string, NULL);
	text_width = XmbTextEscapement(font_set, string,
								   strlen(string));
	good_width = text_width + left_margin + right_margin;

	if (widget_left_margin + text_width + right_margin < width)
		if (width > good_width)
			XtVaSetValues(text,
						  XtNleftMargin, left_margin, XtNwidth, good_width,
						  NULL);
		else
			XtVaSetValues(text,
					   XtNleftMargin, width - text_width - right_margin, NULL);
	else if (width < good_width)
		XtVaSetValues(text, XtNwidth, good_width < max_width ? good_width :
					  max_width, NULL);

	return self;
}

//
//  private methods
//

- makeCaretVisible
{
	Dimension width;
	Position widget_left_margin;
	int resWidth;
	XawTextPosition resPos;
	int resHeight;
	
	XtVaGetValues(text, XtNwidth, &width, XtNleftMargin, &widget_left_margin,
				  NULL);
	XawTextSinkFindDistance(sink, 0, 0,
							XawTextGetInsertionPoint(text),
							&resWidth, &resPos, &resHeight);
	if (widget_left_margin + resWidth > width - right_margin)
		XtVaSetValues(text,
					  XtNleftMargin, width - right_margin - resWidth, NULL);
	else if (widget_left_margin + resWidth < left_margin)
		XtVaSetValues(text, XtNleftMargin, left_margin - resWidth, NULL);

	return self;
}

- setResize:(BOOL)whether
{
	resize = whether;
	return self;
}

- (void)textNextLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// reserved for history
}

- (void)textPreviousLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// reserved for history
}

- (void)textNextPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textPreviousPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textBeginningOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// reserved for history
}

- (void)textEndOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// reserved for history
}

- (void)textScrollOneLineUp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textScrollOneLineDown:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textNewlineAndIndent:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textNewlineAndBackup:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textNewline:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textInsertFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textSearch:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// reserved for history
}

- (void)textFormParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textInsertFileAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textDoSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textDoReplaceAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

- (void)textPopdownSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// do nothing
}

@end

static void HTextEventHandler(Widget w, HistoryText *self, XEvent *ev,
							  Boolean *continue_to_dispatch)
{
	struct HTextStruct { @defs(HistoryText) } *htext;
	
	htext = (struct HTextStruct *)(_eReceiver = self);
	XtDispatchEventToWidget(htext->translator, ev);
}
