// -*- mode: objc; tab-width: 4 -*-

#include <stdlib.h>
#import "IContext.h"
#import "ICManager.h"

#define DEFAULTARRAYSIZE 5
#define NOMOREICID 0			// Let's not use IC id 0.

static ICIDType last_icid = 0;

@implementation ICManager

- init
{
	if ((self = [super init])) {
		if ((ic_array = calloc(DEFAULTARRAYSIZE, sizeof(ic_array[0]))) !=
			NULL)
		{
			// calloc() cleans the memory.
			array_size = DEFAULTARRAYSIZE;
			
			return self;
		}
	}

	// failure
	[self free];				// array_size == 0 so this works.
		
	return nil;
}

- free
{
	int i;

	for (i = 0; i < array_size; i++)
		if (ic_array[i] != nil)
			[ic_array[i] free];

	[super free];

	return nil;
}

- add:(IContext *)anIC
{
	int i;
	ICIDType icid;

	if ((icid = [self generateNewID]) == NOMOREICID)
		return nil;
	if ((i = [self emptyIndex]) == -1) // ic_array[] is full.
		return nil;

	ic_array[i] = [anIC setID:icid];

	return self;
}

- deleteICWithID:(ICIDType)icid
{
	int i;

	for (i = 0; i < array_size; i++)
		if ([ic_array[i] getID] == icid) {
			[ic_array[i] free];
			ic_array[i] = nil;
		}

	return self;
}

- (IContext *)findICWithID:(ICIDType)icid
{
	int i;

	for (i = 0; i < array_size; i++)
		if (ic_array[i] != nil) // This if is actually not needed but...
			if ([ic_array[i] getID] == icid)
				return ic_array[i];

	return nil;
}

//
//  private methods
//

// On success, returns a unique IC id (in this server).
// On failure, returns NOMOREICID.
- (ICIDType)generateNewID
{
	ICIDType new_icid;

	for (new_icid = last_icid+1; new_icid != last_icid;
		 new_icid++, (new_icid == 0) ? new_icid=1 : 0)
		if ([self findICWithID:new_icid] == nil) // new_icid is new!
			break;

	if (new_icid == last_icid)	// IC id space is full.
		return NOMOREICID;

	return last_icid = new_icid;
}

// Finds an empty ic_array[] element and returns its index.
// If the array is full, tries to expand it.
// On failure, returns -1.
- (int)emptyIndex
{
	int i;
	int new_array_size;
	IContext **new_ic_array;
	
	for (i = 0; i < array_size; i++)
		if (ic_array[i] == nil)
			return i;

	// array full. try to expand.
	new_array_size = array_size + array_size/2;
	if ((new_ic_array =
		 realloc(ic_array, new_array_size * sizeof(ic_array[0]))) == NULL)
		return -1;				// array expansion failure
	for (i = array_size; i < new_array_size; i++)
		new_ic_array[i] = nil;
	
	i = array_size;
	array_size = new_array_size;
	ic_array = new_ic_array;

	return i;
}

@end
