// -*- mode: objc; tab-width: 4 -*-

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#import  <objc/objc-api.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "Intrinsic.h"
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>
#include "AdoptedShe.h"

#import  "KAutomataForX.h"
#import  "HistoryText.h"
#import  "CallbackText.h"
#import  "IContext.h"
#import  "byeoroo.h"
#include "hangul.h"

#define MAXMETHODLEN 48
#define MAXLOOKEDUPSTRINGLEN 20
#define MAXCODECONVERTEDTEXTLEN 20
#define SAMPLESTATUSSTRING  " "
#define SAMPLEPREEDITSTRING "ݸ  ӿ  ġ"

#define SET_TEMPLATE(type, member, from) \
{ \
	member = *(type *)from->value; \
}

#define GET_TEMPLATE(type, member, to) \
{ \
	if ((to->value = malloc(sizeof(member))) != NULL) { \
		*(type *)to->value = member; \
		to->value_length = sizeof(member); \
	} \
	else { \
		print_e(0, "Memory allocation failed getting " #member \
				" attribute.\n"); \
		to->value_length = 0; \
	} \
}

// This table associates an input style with a set of preedit and status
// widgets.
static struct widget_table {
	XIMStyle input_style;
	WidgetClass *boxClass;
	WidgetClass *editShellClass;
	enum { NOOBJ, HTEXT, CALLBACKTEXT } textClass;
	WidgetClass *statusShellClass;
	WidgetClass *labelClass;
} widget_classes_by_input_style[] = {
	{ XIMPreeditCallbacks | XIMStatusCallbacks,
	  NULL, NULL, CALLBACKTEXT, NULL, NULL },
    { XIMPreeditCallbacks | XIMStatusArea,
	  NULL, NULL, CALLBACKTEXT, &adoptedShellWidgetClass, &labelWidgetClass },
    { XIMPreeditCallbacks | XIMStatusNothing,
	  NULL, NULL, CALLBACKTEXT, &transientShellWidgetClass,
	  &labelWidgetClass },
    { XIMPreeditCallbacks | XIMStatusNone,
	  NULL, NULL, CALLBACKTEXT, NULL, NULL },
    { XIMPreeditPosition  | XIMStatusCallbacks,
	  NULL, &adoptedShellWidgetClass, HTEXT, NULL, NULL },
    { XIMPreeditPosition  | XIMStatusArea,
	  NULL, &adoptedShellWidgetClass, HTEXT,
	  &adoptedShellWidgetClass, &labelWidgetClass },
    { XIMPreeditPosition  | XIMStatusNothing,
	  NULL, &adoptedShellWidgetClass, HTEXT,
	  &transientShellWidgetClass, &labelWidgetClass },
    { XIMPreeditPosition  | XIMStatusNone,
	  NULL, &adoptedShellWidgetClass, HTEXT, NULL, NULL },
    { XIMPreeditArea      | XIMStatusCallbacks,
	  NULL, &adoptedShellWidgetClass, HTEXT, NULL, NULL },
    { XIMPreeditArea      | XIMStatusArea,
	  NULL, &adoptedShellWidgetClass, HTEXT,
	  &adoptedShellWidgetClass, &labelWidgetClass },
    { XIMPreeditArea      | XIMStatusNothing,
	  NULL, &adoptedShellWidgetClass, HTEXT,
	  &transientShellWidgetClass, &labelWidgetClass },
    { XIMPreeditArea      | XIMStatusNone,
	  NULL, &adoptedShellWidgetClass, HTEXT, NULL, NULL },
    { XIMPreeditNothing   | XIMStatusCallbacks,
	  NULL, &transientShellWidgetClass, HTEXT, NULL, NULL },
    { XIMPreeditNothing   | XIMStatusArea,
	  NULL, &transientShellWidgetClass, HTEXT,
	  &adoptedShellWidgetClass, &labelWidgetClass },
    { XIMPreeditNothing   | XIMStatusNothing,
	  &boxWidgetClass, &transientShellWidgetClass, HTEXT,
	  NULL, &labelWidgetClass },
    { XIMPreeditNothing   | XIMStatusNone,
	  NULL, &transientShellWidgetClass, HTEXT, NULL, NULL },
    { XIMPreeditNone      | XIMStatusCallbacks,
	  NULL, NULL, NOOBJ, NULL, NULL },
    { XIMPreeditNone      | XIMStatusArea,
	  NULL, NULL, NOOBJ, &adoptedShellWidgetClass, &labelWidgetClass },
    { XIMPreeditNone      | XIMStatusNothing,
	  NULL, NULL, NOOBJ, &transientShellWidgetClass, &labelWidgetClass },
    { XIMPreeditNone      | XIMStatusNone,
	  NULL, NULL, NOOBJ, NULL, NULL }
};

static char *find_file(char *dest_path, char *file);
static char *compose_method_name(char *prefix, XICAttribute *attr,
								 char *postfix);

@implementation IContext : Object

- init
{
	[self free];
	return nil;
}

- initWith:(IMChangeICStruct *)data
{
	if ((self = [super init])) {
		char kdf_path[FILENAME_MAX];
		
		connect_id = data->connect_id;
		
		// initialize IC attributes.

		filter_events = NoEventMask; // FILTEREVENTMASK;
		//filter_events = FILTEREVENTMASK;

		// initialize other members.

		line_mode = 0;			// character mode
		if ((kdf_name = strdup(app_data.default_kdf)) == NULL ||
			(ktt_name = strdup("raw.ktt")) == NULL)
		{
			print_e(0, NULL);
			goto fail;
		}
		if (find_file(kdf_path, kdf_name) == NULL) {
			print_e(0, "Cannot find `%s.'\n", kdf_name);
			goto fail;
		}
		if (!(am = [[KAutomataForX alloc] initFromFile:kdf_path on:display]))
			goto fail;
			
		[self setICValues:data];
			
		return self;
	}

fail:
	[self free];
	return nil;
}

- free
{
	if (resource_name != NULL)
		free(resource_name);
	if (resource_class != NULL)
		free(resource_class);

	if (kdf_name)
		free(kdf_name);
	if (ktt_name)
		free(ktt_name);
	[am free];
	
	// The destroyal of text must precede that of editShell because the latter
	// also destroys text->text.
	[text free];
	if (editShell)
		XtDestroyWidget(editShell);
	if (statusShell)
		XtDestroyWidget(statusShell);
	
	[super free];

	return nil;
}

- setID:(ICIDType)newIcid
{
	icid = newIcid;
	return self;
}

- (ICIDType)getID
{
	return icid;
}

- changeParameters
{
	char *new_kdf_name, *new_ktt_name;
	BOOL new_line_mode;
	char path[FILENAME_MAX];
	BOOL events_forwarded = line_mode || keyboard != MRAW;
	
	XtVaGetValues(cpkdf, XtNlabel, &new_kdf_name, NULL);
	XtVaGetValues(cpktt, XtNlabel, &new_ktt_name, NULL);
	new_line_mode = strcmp(XawToggleGetCurrent(cpchar), "cpline") == 0;

	if (strcmp(kdf_name, new_kdf_name) != 0) {
		KAutomataForX *new_am;
		
		if (find_file(path, new_kdf_name) == NULL) {
			print_e(0, "Cannot find `%s.'\n", new_kdf_name);
			goto kdf_end;
		}
		if (!(new_am = [[KAutomataForX alloc] initFromFile:path on:display])) {
			print_e(0, "Cannot initialize Korean automata.\n");
			goto kdf_end;
		}

		if (text && ![text setKDF:path]) {
			[new_am free];
			goto kdf_end;
		}

		free(kdf_name);
		kdf_name = strdup(new_kdf_name);
		[am free];
		am = new_am;
	}
kdf_end:
	if (strcmp(ktt_name, new_ktt_name) != 0) {
		free(ktt_name);
		ktt_name = strdup(new_ktt_name);
	}
//ktt_end:
	if (!!new_line_mode != !!line_mode) {
		line_mode = new_line_mode;
		[[self flushAutomata] flushText];
		if (line_mode) {
			[text setKeyboard:keyboard];
			if (input_style & XIMPreeditPosition) {
				[self changeTextMaxWidth];
				[[[text setLeftMargin:2] setRightMargin:2] setResize:YES];
			}
		}
		else {
			// We don't have to keyboard = [text getKeyboard] if !line_mode
			// because we already did that enough in -forwardEvent:.
			if (input_style & XIMPreeditPosition) {
				[self setCharModeTextWidth];
				[[[text setLeftMargin:0] setRightMargin:0] setResize:NO];
			}
		}
	}

	if (events_forwarded && (!line_mode && keyboard == MRAW))
		[self endPreedit];
	else if (!events_forwarded && (line_mode || keyboard != MRAW))
		[self startPreedit];
		
	[self changeStatusLabel];
	return self;
}

//
//  XIM protocol methods
//

- (int)setICFocus:(IMChangeFocusStruct *)data
{
	focusedIC = self;

	[self changeStatusLabel];
	XtVaSetValues(kdfLabel, XtNlabel, kdf_name, NULL);
	XtVaSetValues(kttLabel, XtNlabel, ktt_name, NULL);
	XtSetSensitive(box, True);
	if (statusShell)
		XtSetSensitive(statusShell, True);

	if (input_style & XIMPreeditArea)
		XtPopup(editShell, XtGrabNone);

	if (input_style & XIMStatusArea)
		XtPopup(statusShell, XtGrabNone);

	return True;
}

- (int)unsetICFocus:(IMChangeFocusStruct *)data
{
	focusedIC = nil;
	XtSetSensitive(box, False);
	if (statusShell)
		XtSetSensitive(statusShell, False);
	
	return True;
}

- (int)resetIC:(IMResetICStruct *)data
{
	return True;
}

- (int)setICValues:(IMChangeICStruct *)data
{
	int num_ic_attr = data->ic_attr_num;
	int num_p_attr = data->preedit_attr_num;
	int num_s_attr = data->status_attr_num;
	XICAttribute *ic_attr = data->ic_attr;
	XICAttribute *p_attr = data->preedit_attr;
	XICAttribute *s_attr = data->status_attr;
	int i;
	SEL sel;
	
	for (i = 0; i < num_ic_attr; i++) {
		sel = sel_get_uid(compose_method_name("setN", &ic_attr[i], ":"));
		if ([self respondsTo:sel])
			[self perform:sel with:(id)&ic_attr[i]];
#ifdef DEBUG
		else
			print_e(0, "Method -%s unused.\n", sel_get_name(sel));
#endif
	}
	for (i = 0; i < num_p_attr; i++) {
		sel = sel_get_uid(compose_method_name("setN", &p_attr[i], ":of:"));
		if ([self respondsTo:sel])
			[self perform:sel with:(id)&p_attr[i] with:(id)&preedit_attr];
#ifdef DEBUG
		else
			print_e(0, "Method -%s unused.\n", sel_get_name(sel));
#endif
	}
	for (i = 0; i < num_s_attr; i++) {
		sel = sel_get_uid(compose_method_name("setN", &s_attr[i], ":of:"));
		if ([self respondsTo:sel])
			[self perform:sel with:(id)&s_attr[i] with:(id)&status_attr];
#ifdef DEBUG
		else
			print_e(0, "Method -%s unused.\n", sel_get_name(sel));
#endif
	}
	
	return True;
}

- (int)getICValues:(IMChangeICStruct *)data
{
	int num_ic_attr = data->ic_attr_num;
	int num_p_attr = data->preedit_attr_num;
	int num_s_attr = data->status_attr_num;
	XICAttribute *ic_attr = data->ic_attr;
	XICAttribute *p_attr = data->preedit_attr;
	XICAttribute *s_attr = data->status_attr;
	int i;
	SEL sel;
	
	for (i = 0; i < num_ic_attr; i++) {
		sel = sel_get_uid(compose_method_name("getN", &ic_attr[i], ":"));
		if ([self respondsTo:sel])
			[self perform:sel with:(id)&ic_attr[i]];
#ifdef DEBUG
		else
			print_e(0, "Method -%s unused.\n", sel_get_name(sel));
#endif
	}
	for (i = 0; i < num_p_attr; i++) {
		sel = sel_get_uid(compose_method_name("getN", &p_attr[i], ":of:"));
		if ([self respondsTo:sel])
			[self perform:sel with:(id)&p_attr[i] with:(id)&preedit_attr];
#ifdef DEBUG
		else
			print_e(0, "Method -%s unused.\n", sel_get_name(sel));
#endif
	}
	for (i = 0; i < num_s_attr; i++) {
		sel = sel_get_uid(compose_method_name("getN", &s_attr[i], ":of:"));
		if ([self respondsTo:sel])
			[self perform:sel with:(id)&s_attr[i] with:(id)&status_attr];
#ifdef DEBUG
		else
			print_e(0, "Method -%s unused.\n", sel_get_name(sel));
#endif
	}
	
	return True;
}

// We use dynamic event flow.
// There are no off keys registered and hence this method will be
// called only when preedit begins.
// We can assume here: !line_mode && keyboard == MRAW.
- (int)triggerNotify:(IMTriggerNotifyStruct *)data
{
	int key_index = data->key_index / 3;

	switch (key_index) {
	case TOGGLE:
		keyboard = MKOREAN;
		[self changeStatusLabel];
		break;
	case CONVERSION:
		print_e(0, "Chinese conversion!\n");
		break;
	case POPUP:
		[self popupControlPanel:(XPointer)data];
		break;
#ifdef DEBUG
	default:
		print_e(0, "Strange trigger key, %d.\n", data->key_index);
		IMPreeditEnd(ims, (XPointer)data);
		return False;
#endif
	}

	if (!line_mode && keyboard == MRAW)
		IMPreeditEnd(ims, (XPointer)data);
	return True;
}

- (int)forwardEvent:(IMForwardEventStruct *)data
{
	XEvent *ev = &data->event;
	
	if (ev->type == KeyPress) {
		int key_index;

		if ((key_index = trigger_key_index(&ev->xkey)) >= 0) { // a trigger key
			switch(key_index) {
			case TOGGLE:
				if (line_mode) {
					[text receiveEvent:ev];
					keyboard = [text getKeyboard];
				}
				else {
					[self flushAutomata];
					keyboard = MRAW;
				}
				[self changeStatusLabel];
				break;
			case POPUP:
				[self popupControlPanel:(XPointer)data];
				break;
			case CONVERSION:
				print_e(0, "Chinese conversion!\n");
				break;
			}
		}
		else {					// not a trigger key
			if (line_mode) {
				char buf[MAXLOOKEDUPSTRINGLEN];
				KeySym keysym;

				buf[XLookupString(&ev->xkey, buf, sizeof(buf), &keysym, NULL)]
					= '\0';
				if (keysym == XK_Return) {
					[self flushText];
					IMForwardEvent(ims, (XPointer)data);
				}
				else {
					BOOL text_was_empty, text_is_empty;
					
					if (input_style & XIMPreeditPosition) {
						char *string;

						XtVaGetValues(preedit_attr.widget,
									  XtNstring, &string, NULL);
						text_was_empty = string[0] == '\0';
					}

					[text receiveEvent:ev];
					
					if (input_style & XIMPreeditPosition) {
						char *string;
						
						XtVaGetValues(preedit_attr.widget,
									  XtNstring, &string, NULL);
						text_is_empty = string[0] == '\0';
						if (text_was_empty && !text_is_empty)
							XtPopup(editShell, XtGrabNone);
						else if (!text_was_empty && text_is_empty)
							XtPopdown(editShell);
						else if (text_was_empty && text_is_empty)
							IMForwardEvent(ims, (XPointer)data);
					}
				}
			} // line_mode
			else
				if (!is_modifier(ev->xkey.keycode))
					[self driveAutomata:data];

			if (keyboard == MKOREAN)
				XtVaSetValues(korEngStat,
							  XtNlabel, [self getComposingChar], NULL);
		} // not a trigger key

		if (input_style & XIMPreeditPosition) {
			if (!line_mode)
				if ([am getComposingChar][0])
					XtPopup(editShell, XtGrabNone);
				else
					XtPopdown(editShell);
		}
	}

	if (!line_mode && keyboard == MRAW)
		[self endPreedit];
		
	return True;
}

- (int)preeditStartReply:(IMPreeditCBStruct *)data
{
	return True;
}

- (int)preeditCaretReply:(IMPreeditCBStruct *)data
{
	return True;
}

- (int)strConversionReply:(IMStrConvCBStruct *)data
{
	return True;
}

//
//  private methods
//

- startPreedit
{
	IMPreeditStateStruct ps;
	
	ps.connect_id = connect_id;
	ps.icid = icid;
	IMPreeditStart(ims, (XPointer)&ps);
	return self;
}

- endPreedit
{
	IMPreeditStateStruct ps;
	
	ps.connect_id = connect_id;
	ps.icid = icid;
	IMPreeditEnd(ims, (XPointer)&ps);
	return self;
}

- changeStatusLabel
{
	char *mode;
	char status_string[5];
	
	switch (keyboard) {
	case MRAW:
		XtVaSetValues(korEngStat, XtNlabel, "A", NULL);
		mode = "";
		break;
	case MTRANS:
		XtVaSetValues(korEngStat, XtNlabel, "Ư", NULL);
		mode = "Ư";
		break;
	case MKOREAN:
		XtVaSetValues(korEngStat, XtNlabel, [self getComposingChar], NULL);
		mode = "";
		break;
	}
	XtVaSetValues(charLineStat, XtNlabel, line_mode ? "" : "", NULL);
	sprintf(status_string, "%s%s", mode, line_mode ? "" : "");
	if (status_attr.widget)
		XtVaSetValues(status_attr.widget, XtNlabel, status_string, NULL);

	return self;
}

- (char *)getComposingChar
{
	// line_mode == 0 always in XIMPreeditNone style.  Hence no problem in
	// the next line.
	return [line_mode ? text : am getComposingChar];
}

- driveAutomata:(IMForwardEventStruct *)data
{
	char wansung[MAXCODECONVERTEDTEXTLEN];
	char buf[MAXLOOKEDUPSTRINGLEN];
	KeySym keysym;
	XKeyEvent *kev = &data->event.xkey;
	
	buf[XLookupString(kev, buf, sizeof(buf), &keysym, NULL)] = '\0';
	
	if (keysym == XK_BackSpace || keysym == XK_Delete) {
		if ([am deleteLastKey] == EMPTY)
			IMForwardEvent(ims, (XPointer)data);
	}
	else
		switch ([am eatKeycode:kev->keycode modifier:kev->state]) {
		case CHARBREAK:
			[self commitJohabString:[am getComposedChar]];
		case JUSTFINE:
			break;
		case WRONGKEY:
			[self flushAutomata];
			IMForwardEvent(ims, (XPointer)data);
			break;
#ifdef DEBUG
		default:
			print_e(0, "automata error.\n");
#endif
		}

	if (!(input_style & XIMPreeditNone)) {
		wansung[convert_johab_to_ks([am getComposingChar],
							   wansung, strlen([am getComposingChar]))] = '\0';
		[text setString:wansung];
	}
	return self;
}

- flushAutomata
{
	if ([am getComposingChar][0]) {
		[self commitJohabString:[am getComposingChar]];
		[am reset];
		[text setString:""];
	}
	return self;
}

- flushText
{
	if ([text getString][0]) {
		[self commitString:[text getString]];
		[text setString:""];
	}
	if (input_style & XIMPreeditPosition)
		XtPopdown(editShell);
	return self;
}

- commitString:(char *)wansung_string
{
	IMCommitStruct cs;
	XTextProperty tp;
	
	XmbTextListToTextProperty(display, (char **)&wansung_string, 1,
							  XCompoundTextStyle, &tp);
	cs.connect_id = connect_id;
	cs.icid = icid;
	cs.flag = XimLookupChars;
	cs.commit_string = tp.value;
	IMCommitString(ims, (XPointer)&cs);
	XFree(tp.value);

	return self;
}

- commitJohabString:(char *)johab_string
{
	char wansung[MAXCODECONVERTEDTEXTLEN];

	wansung[convert_johab_to_ks(johab_string, wansung, strlen(johab_string))]
		= '\0';
	return [self commitString:wansung];
}

- commitKeysym:(KeySym)keysym
{
	IMCommitStruct cs;

	cs.connect_id = connect_id;
	cs.icid = icid;
	cs.flag = XimLookupKeySym;
	cs.commit_string = "";
	cs.keysym = keysym;
	IMCommitString(ims, (XPointer)&cs);

	return self;
}

- setAlreadySetAttributes
{
	XICAttribute attr;
	
	if (client_window) {
		attr.value = &client_window;
		[self setNclientWindow:&attr];
	}

	return self;
}

- setAlreadySetSubattributesOf:(SubAttr *)subattr
{
	XICAttribute attr;
		
	if (subattr->area.width != 0 || subattr->area.height != 0) {
		attr.value = &subattr->area;
		[self setNarea:&attr of:subattr];
	}
	if (subattr->colormap) {
		attr.value = &subattr->colormap;
		[self setNcolorMap:&attr of:subattr];
	}
	if (subattr->std_colormap) {
		attr.value = &subattr->std_colormap;
		[self setNstdColorMap:&attr of:subattr];
	}
	if (subattr->foreground_set) {
		attr.value = &subattr->foreground;
		[self setNforeground:&attr of:subattr];
	}
	if (subattr->background_set) {
		attr.value = &subattr->background;
		[self setNbackground:&attr of:subattr];
	}
	if (subattr->background_pixmap) { // FIXME!
		attr.value = &subattr->background_pixmap;
		[self setNbackgroundPixmap:&attr of:subattr];
	}
	if (subattr->font_set)
		XtVaSetValues(subattr->widget,
					  XtNfontSet, subattr->font_set, NULL);
	if (subattr->cursor) {
		attr.value = &subattr->cursor;
		[self setNcursor:&attr of:subattr];
	}

	return self;
}

- popupControlPanel:(XPointer)data
{
	Window root, child;
	int rx, ry, wx, wy;
	unsigned int mask;
	
	if (!cp_popped_up) {
		lastFocusedIC = focusedIC;
	
		// There is no line mode in XIMPreeditNone style.
		XtSetSensitive(cpline, !(input_style & XIMPreeditNone));

		if (line_mode)
			XtVaSetValues(cpline, XtNstate, True, NULL);
		else
			XtVaSetValues(cpchar, XtNstate, True, NULL);
		XtVaSetValues(cpkdf, XtNlabel, kdf_name, NULL);
		XtVaSetValues(cpktt, XtNlabel, ktt_name, NULL);
	
		XQueryPointer(display, XtWindow(topLevel),
					  &root, &child, &rx, &ry, &wx, &wy, &mask);
		XtVaSetValues(cpshell,
					  XtNx, rx - 70, XtNy, ry - 70, NULL);
		XtPopup(cpshell, XtGrabNonexclusive);
		cp_popped_up = 1;
	}
	else
		XBell(display, 0);
	
	return self;
}

- changeTextMaxWidth
{
	if (line_mode) {
		Position area_right_edge = preedit_attr.area.x +
			                       preedit_attr.area.width;
		Position text_left_edge = preedit_attr.spot_location.x +
			                      preedit_attr.max_logical_extent.x - 2;
		
		[text setMaxWidth:area_right_edge - text_left_edge];
	}
	return self;
}

- setCharModeTextWidth
{
	XtVaSetValues(preedit_attr.widget,
				  XtNwidth, preedit_attr.max_logical_extent.width+2, NULL);
	[text setMaxWidth:preedit_attr.max_logical_extent.width+2];
	return self;
}

//
//  IC attribute methods
//

// XNInputStyle is set only once when the IC is created, and we cannot
// decide which class of widgets we have to create before this attribute
// is set.  So, we create the widgets here.
- setNinputStyle:(XICAttribute *)attr
{
	struct widget_table *p;
	Widget new_ebox = None;
	Widget new_editShell = None;
	id new_text = nil;
	Widget new_statusShell = None;
	Widget new_status = None;
	
	SET_TEMPLATE(XIMStyle, input_style, attr);

	// look up widget classes by input style.
	for (p = widget_classes_by_input_style;
		 p < widget_classes_by_input_style +
			 XtNumber(widget_classes_by_input_style);
		 p++)
		if (p->input_style == input_style)
			break;
	if (p == widget_classes_by_input_style +
		XtNumber(widget_classes_by_input_style))
	{
		print_e(0, "Unsupported input style: 0x%x.\n", input_style);
		return nil;
	}
	
	if (p->editShellClass)
		if (!(new_editShell = XtVaCreatePopupShell("editShell",
							  *p->editShellClass, topLevel, NULL)))
			goto fail;
	if (p->statusShellClass)
		if (!(new_statusShell = XtVaCreatePopupShell("statusShell",
								*p->statusShellClass, topLevel, NULL)))
			goto fail;
	if (p->boxClass)
		if (!(new_ebox = XtVaCreateManagedWidget("editBox", *p->boxClass,
												 new_editShell, NULL)))
			goto fail;
	if (p->labelClass)
		if (!(new_status = XtVaCreateManagedWidget("status", *p->labelClass,
						   new_ebox ? new_ebox : new_statusShell, NULL)))
			goto fail;
	if (p->textClass != NOOBJ) {
		char kdf_path[FILENAME_MAX];
		char ktt_path[FILENAME_MAX];
		
		if (find_file(kdf_path, kdf_name) == NULL) {
			print_e(0, "Cannot find `%s.'\n", kdf_name);
			goto fail;
		}
		if (find_file(ktt_path, ktt_name) == NULL) {
			print_e(0, "Cannot find `%s.'\n", ktt_name);
			goto fail;
		}
		switch (p->textClass) {
			Widget asciiText;

		case HTEXT:
			if (!(asciiText = XtVaCreateManagedWidget("preedit",
							  asciiTextWidgetClass, new_ebox ? new_ebox :
							  new_editShell, NULL)))
				goto fail;
			new_text = [[HistoryText alloc] initFromKDF:kdf_path
									withWidget:asciiText translator:topLevel];
			if (!new_text)
				goto fail;
			if (XtClass(XtParent(asciiText)) == adoptedShellWidgetClass) {
				XtVaSetValues(asciiText,
							  XtNtopMargin, 0, XtNbottomMargin, 0, NULL);
				if (input_style & XIMPreeditPosition)
					[[[new_text setLeftMargin:0] setRightMargin:0]
						                         setResize:NO];
			}
			break;
		case CALLBACKTEXT:
			if (!(new_text = [[CallbackText alloc] init]))
				goto fail;
			break;
		}
	}
	
	if (new_ebox)        ebox        = new_ebox;
	if (new_editShell)   editShell   = new_editShell;
	if (new_statusShell) statusShell = new_statusShell;
	if (new_status)      status_attr.widget = status = new_status;
	if (new_text)        preedit_attr.widget = [text = new_text getWidget];
		
	// cause preedit_attr.widget to work in character mode.
// 	if (preedit_attr.widget)
// 		XtInsertEventHandler(preedit_attr.widget,
// 							 [text getTextEventMask], False,
// 							 (XtEventHandler)CharModeTextEventHandler,
// 							 self, XtListHead);

	// There may be IC attributes already set.
	[[[self setAlreadySetAttributes]
		    setAlreadySetSubattributesOf:&preedit_attr]
		    setAlreadySetSubattributesOf:&status_attr];
	
	return self;

fail:
	[new_text free];			// This must precede the destroyal of
								// new_editShell because it also destroys
								// new_text->text.
	if (new_editShell)
		XtDestroyWidget(new_editShell);
	if (new_statusShell)
		XtDestroyWidget(new_statusShell);
	return nil;
}

- getNinputStyle:(XICAttribute *)attr
{
	GET_TEMPLATE(XIMStyle, input_style, attr);
	return self;
}

- setNclientWindow:(XICAttribute *)attr
{
	SET_TEMPLATE(Window, client_window, attr);
	if (!focus_window)
		focus_window = client_window;
	
	if (editShell && XtClass(editShell) == adoptedShellWidgetClass)
		XtVaSetValues(editShell,
					  XtNparentWindow, (input_style & XIMPreeditPosition) ?
					  focus_window : client_window, NULL);
	if (statusShell && XtClass(statusShell) == adoptedShellWidgetClass)
		XtVaSetValues(statusShell, XtNparentWindow, client_window, NULL);
	
	return self;
}

- getNclientWindow:(XICAttribute *)attr
{
	GET_TEMPLATE(Window, client_window, attr);
	return self;
}

- setNfocusWindow:(XICAttribute *)attr
{
	SET_TEMPLATE(Window, focus_window, attr);
	if (input_style & XIMPreeditPosition)
		XtVaSetValues(editShell, XtNparentWindow, focus_window, NULL);
		
	return self;
}

- getNfocusWindow:(XICAttribute *)attr
{
	GET_TEMPLATE(Window, focus_window, attr);
	return self;
}

- setNresourceName:(XICAttribute *)attr
{
	// I'm not sure if this treatment is right.
	char *new_resource_name;
	
	// attr->value may not be terminated by '\0'.
	if ((new_resource_name = malloc(attr->value_length+1)) != NULL) {
		memcpy(new_resource_name, attr->value, attr->value_length);
		new_resource_name[attr->value_length] = '\0';
		if (resource_name != NULL)
			free(resource_name);
		resource_name = new_resource_name;
	}
	return self;
}

- getNresourceName:(XICAttribute *)attr
{
	if ((attr->value = strdup(resource_name)) != NULL)
		attr->value_length = strlen(resource_name);
	else {
		print_e(0,
				"Memory allocation failed getting resource_name attribute.\n");
		attr->value_length = 0;
	}
	return self;
}

- setNresourceClass:(XICAttribute *)attr
{
	// I'm not sure if this treatment is right.
	char *new_resource_class;
	
	// attr->value may not be terminated by '\0'.
	if ((new_resource_class = malloc(attr->value_length+1)) != NULL) {
		memcpy(new_resource_class, attr->value, attr->value_length);
		new_resource_class[attr->value_length] = '\0';
		if (resource_class != NULL)
			free(resource_class);
		resource_class = new_resource_class;
	}
	return self;
}

- getNresourceClass:(XICAttribute *)attr
{
	if ((attr->value = strdup(resource_class)) != NULL)
		attr->value_length = strlen(resource_class);
	else {
		print_e(0, "Memory alloc. failed getting resource_class attribute.\n");
		attr->value_length = 0;
	}
	return self;
}

- setNfilterEvents:(XICAttribute *)attr
{
	print_e(0, "-setNfilterEvents: must not be called at all!\n");
	return self;
}

- getNfilterEvents:(XICAttribute *)attr
{
	GET_TEMPLATE(EventMask, filter_events, attr);
	return self;
}

//
//  preedit & status sub-attribute methods
//

- setNarea:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(XRectangle, subattr->area, attr);

	if (subattr == &preedit_attr) {
		if (input_style & XIMPreeditPosition)
			[self changeTextMaxWidth];
		else if (input_style & XIMPreeditArea) {
			XtVaSetValues(subattr->widget,
						  XtNwidth, subattr->area.width,
						  XtNheight, subattr->area.height, NULL);
			XtVaSetValues(editShell,
						  XtNx, subattr->area.x,
						  XtNy, subattr->area.y - 2, NULL);
		}
	}
	else {						// subattr == &status_attr
		if (input_style & XIMStatusArea) {
			XtVaSetValues(subattr->widget,
						  XtNwidth, subattr->area.width,
						  XtNheight, subattr->area.height, NULL);
			XtVaSetValues(statusShell,
						  XtNx, subattr->area.x,
						  XtNy, subattr->area.y - 2, NULL);
		}
	}

	// reset geometry negotiation.
	subattr->area_needed.width = subattr->area_needed.height = 0;
	
	return self;
}

- getNarea:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(XRectangle, subattr->area, attr);
	return self;
}

- setNareaNeeded:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(XRectangle, subattr->area_needed, attr);
#ifdef DEBUG
	print_e(0, "-setNareaNeeded: area_needed = (%d, %d, %d, %d)\n",
			subattr->area_needed.x, subattr->area_needed.y,
			subattr->area_needed.width, subattr->area_needed.height);
#endif
	return self;
}

// Compute my desired size based on area_needed and return it to the client.
- getNareaNeeded:(XICAttribute *)attr of:(SubAttr *)subattr
{
	XRectangle my_desired_area = { 0, 0, 0, 0 };
	XRectangle *area_needed = &subattr->area_needed;
	
	my_desired_area.height = subattr->max_logical_extent.height + 2;
	
	if (subattr == &status_attr) {
		my_desired_area.width =
			XmbTextEscapement(subattr->font_set, SAMPLESTATUSSTRING,
							  sizeof(SAMPLESTATUSSTRING) - 1);
		if (area_needed->width == 0 && area_needed->height != 0)
			my_desired_area.height = area_needed->height;
		else if (area_needed->width != 0 && area_needed->height == 0)
			if (area_needed->width >= my_desired_area.width)
				my_desired_area.width = area_needed->width;
			else {
				my_desired_area.height += my_desired_area.height *
					(my_desired_area.width-1)/area_needed->width;
				my_desired_area.width = area_needed->width;
			}
		else if (area_needed->width != 0 && area_needed->height != 0) {
			my_desired_area.width = area_needed->width;
			my_desired_area.height = area_needed->height;
		}
		// We have nothing to do if both width and height are not constrained.
	}
	else {						// preedit area negotiation.
		my_desired_area.width =
			XmbTextEscapement(subattr->font_set, SAMPLEPREEDITSTRING,
							  sizeof(SAMPLEPREEDITSTRING) - 1);
		if (area_needed->width == 0 && area_needed->height != 0)
			my_desired_area.height = area_needed->height;
		else if (area_needed->width != 0 && area_needed->height == 0)
			my_desired_area.width = area_needed->width;
		else if (area_needed->width != 0 && area_needed->height != 0) {
			my_desired_area.width = area_needed->width;
			my_desired_area.height = area_needed->height;
		}
	}
	
	GET_TEMPLATE(XRectangle, my_desired_area, attr);
	return self;
}

- setNspotLocation:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(XPoint, subattr->spot_location, attr);

	if (subattr == &preedit_attr)
		if (input_style & XIMPreeditPosition) {
			[self changeTextMaxWidth];
			XtVaSetValues(editShell,
						  XtNx, subattr->spot_location.x +
						  subattr->max_logical_extent.x - 2,
						  XtNy, subattr->spot_location.y +
						  subattr->max_logical_extent.y - 2,
						  NULL);
#ifdef DEBUG
			print_e(0, "New spot_location = (%d, %d)\n",
					subattr->spot_location.x, subattr->spot_location.y);
#endif
		}
	return self;
}

- getNspotLocation:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(XPoint, subattr->spot_location, attr);
	return self;
}

- setNcolorMap:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(Colormap, subattr->colormap, attr);

	if (subattr->widget)
		XtVaSetValues(subattr->widget, XtNcolormap, subattr->colormap, NULL);
	
	return self;
}

- getNcolorMap:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(Colormap, subattr->colormap, attr);
	return self;
}

- setNstdColorMap:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(Atom, subattr->std_colormap, attr);
	// FIXME!
#ifdef DEBUG
	print_e(0, "-[IContext setNstdColorMap:of:]\n");
#endif
	return self;
}

- getNstdColorMap:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(Atom, subattr->std_colormap, attr);
	// FIXME!
	return self;
}

- setNforeground:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(unsigned long, subattr->foreground, attr);
	subattr->foreground_set = YES;

	if (subattr->widget)
		XtVaSetValues(subattr->widget,
					  XtNforeground, subattr->foreground, NULL);
	
	return self;
}

- getNforeground:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(unsigned long, subattr->foreground, attr);
	return self;
}

- setNbackground:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(unsigned long, subattr->background, attr);
	subattr->background_set = YES;

	if (subattr->widget)
		XtVaSetValues(subattr->widget,
					  XtNbackground, subattr->background, NULL);
	
	return self;
}

- getNbackground:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(unsigned long, subattr->background, attr);
	return self;
}

- setNbackgroundPixmap:(XICAttribute *)attr of:(SubAttr *)subattr
{
	// FIXME!
	print_e(0, "-setNbackgroundPixmap: value = %s.\n", attr->value);
	return self;
}

- getNbackgroundPixmap:(XICAttribute *)attr of:(SubAttr *)subattr
{
	// FIXME!
	return self;
}

- setNfontSet:(XICAttribute *)attr of:(SubAttr *)subattr
{
	XOC new_oc;
	char *base_font_name;
	XOMCharSetList charset_list;

	if ((base_font_name = malloc(attr->value_length+1)) == NULL) {
		print_e(0, "Memory allocation failed setting font_set attribute.\n");
		return self;
	}
	memcpy(base_font_name, attr->value, attr->value_length);
	base_font_name[attr->value_length] = '\0';

	new_oc = XCreateOC(om, XNBaseFontName, base_font_name, NULL);
	free(base_font_name);
	if (new_oc == NULL) {
		print_e(0, "Cannot create output context.\n");
		return self;
	}
	
	if (XGetOCValues(new_oc, XNMissingCharSet, &charset_list, NULL) == NULL) {
		// NULL means okay.
		int i;
		
		for (i = 0; i < charset_list.charset_count; i++)
			print_e(0, "Warning: charset missing: %s\n",
					charset_list.charset_list[i]);

		XFreeStringList(charset_list.charset_list);
	}

	if (subattr->font_set != NULL)
		XDestroyOC(subattr->font_set);
	subattr->font_set = new_oc;
	subattr->max_logical_extent =
		XExtentsOfFontSet(subattr->font_set)->max_logical_extent;

	if (subattr->widget)
		XtVaSetValues(subattr->widget, XtNfontSet, subattr->font_set, NULL);
	if (input_style & XIMPreeditPosition) {
		XtVaSetValues(subattr->widget,
					  XtNheight, subattr->max_logical_extent.height+2, NULL);
		[self setCharModeTextWidth];
	}

	return self;
}

- getNfontSet:(XICAttribute *)attr of:(SubAttr *)subattr
{
	CARD16 list_len;
	int total_len;
	char *base_font_name_list = XBaseFontNameListOfFontSet(subattr->font_set);
	
	if (base_font_name_list == NULL) {
		print_e(0, "XBaseFontNameListOfFontSet() failed.\n");
		goto fail;
	}
	list_len = strlen(base_font_name_list);
	total_len = sizeof(list_len) + list_len;
	if ((attr->value = malloc(total_len)) == NULL) {
		perror("-[IContext getNfontSet:of:]");
		goto fail;
	}
	memcpy(attr->value, &list_len, sizeof(list_len));
	strncpy(attr->value + sizeof(list_len), base_font_name_list, list_len);
	attr->value_length = total_len;

	return self;

fail:
	attr->value_length = 0;
	return self;
}

- setNlineSpace:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(int, subattr->line_space, attr);
	return self;
}

- getNlineSpace:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(int, subattr->line_space, attr);
	return self;
}

- setNcursor:(XICAttribute *)attr of:(SubAttr *)subattr
{
	SET_TEMPLATE(Cursor, subattr->cursor, attr);

	if (subattr->widget)
		XtVaSetValues(subattr->widget, XtNcursor, subattr->cursor, NULL);
	
	return self;
}

- getNcursor:(XICAttribute *)attr of:(SubAttr *)subattr
{
	GET_TEMPLATE(Cursor, subattr->cursor, attr);
	return self;
}

@end

static char *find_file(char *dest_path, char *file)
{	
	char *p, *q;
	
	for (p = app_data.search_path; *p; p = (*q ? q + 1 : q)) {
		if ((q = strchr(p, ':')) == NULL)
			q = p + strlen(p);
		memcpy(dest_path, p, q - p);
		dest_path[q - p] = '/';
		dest_path[q - p + 1] = '\0';
		strcat(dest_path, file);

		if (access(dest_path, R_OK) == 0) // can read
			return dest_path;
	}

	return NULL;
}

// This routine assumes that the string values of XN... name constants
// will not change from version to version of Xlib.
// The method names in this program are as of version 11 release 6.3.
static char *compose_method_name(char *prefix, XICAttribute *attr,
								 char *postfix)
{
	static char return_value[MAXMETHODLEN];
	char buf[MAXMETHODLEN];
	
	// attr->name may not be terminated by '\0'.
	memcpy(buf, attr->name, attr->name_length);
	buf[attr->name_length] = '\0';
	sprintf(return_value, "%s%s%s", prefix, buf, postfix);

	return return_value;
}
