/* -*- mode:c; tab-width: 8 -*- */

/* $XConsortium: Label.c,v 1.97 94/04/17 20:12:12 kaleb Exp $ */

/***********************************************************

Copyright (c) 1987, 1988, 1994  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.


Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/*
 *  Modified by Park, Jae-hyon to display Korean string in
 *  2 byte johab encoding with johab fonts.
 */

/*
 * JohabLabel.c - JohabLabel widget
 */

#include <stdio.h>
#include <ctype.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xos.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xmu/Converters.h>
#include <X11/Xmu/Drawing.h>

#include "JohabLabelP.h"
#include "XJohab.h"

/* needed for abs() */
#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#else
int abs();
#endif

#define streq(a,b) (strcmp( (a), (b) ) == 0)

#define MULTI_LINE_LABEL 32767

#ifdef CRAY
#define WORD64
#endif

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

/* Private Data */

#define offset(field) XtOffsetOf(JohabLabelRec, field)
static XtResource resources[] = {
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(johabLabel.foreground), XtRString, XtDefaultForeground},
    {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(johabLabel.font),XtRString, XtDefaultFont},
    {XtNjohabFont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(johabLabel.johabFont),XtRString, XtDefaultFont},
    {XtNlabel,  XtCLabel, XtRString, sizeof(String),
	offset(johabLabel.label), XtRString, NULL},
    {XtNjustify, XtCJustify, XtRJustify, sizeof(XtJustify),
	offset(johabLabel.justify), XtRImmediate, (XtPointer)XtJustifyCenter},
    {XtNinternalWidth, XtCWidth, XtRDimension,  sizeof(Dimension),
	offset(johabLabel.internal_width), XtRImmediate, (XtPointer)4},
    {XtNinternalHeight, XtCHeight, XtRDimension, sizeof(Dimension),
	offset(johabLabel.internal_height), XtRImmediate, (XtPointer)2},
    {XtNresize, XtCResize, XtRBoolean, sizeof(Boolean),
	offset(johabLabel.resize), XtRImmediate, (XtPointer)True},
};
#undef offset

static void Initialize();
static void Resize();
static void Redisplay();
static Boolean SetValues();
static void ClassInitialize();
static void Destroy();
static XtGeometryResult QueryGeometry();

JohabLabelClassRec johabLabelClassRec = {
  {
/* core_class fields */	
    /* superclass	  	*/	(WidgetClass) &simpleClassRec,
    /* class_name	  	*/	"JohabLabel",
    /* widget_size	  	*/	sizeof(JohabLabelRec),
    /* class_initialize   	*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited       	*/	FALSE,
    /* initialize	  	*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize		  	*/	XtInheritRealize,
    /* actions		  	*/	NULL,
    /* num_actions	  	*/	0,
    /* resources	  	*/	resources,
    /* num_resources	  	*/	XtNumber(resources),
    /* xrm_class	  	*/	NULLQUARK,
    /* compress_motion	  	*/	TRUE,
    /* compress_exposure  	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest	  	*/	FALSE,
    /* destroy		  	*/	Destroy,
    /* resize		  	*/	Resize,
    /* expose		  	*/	Redisplay,
    /* set_values	  	*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus	 	*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private   	*/	NULL,
    /* tm_table		   	*/	NULL,
    /* query_geometry		*/	QueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
/* Simple class fields initialization */
  {
    /* change_sensitive		*/	XtInheritChangeSensitive
  },
/* Label class fields initialization */
  {
    /* ignore 			*/	0
  }
};
WidgetClass johabLabelWidgetClass = (WidgetClass)&johabLabelClassRec;
/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void ClassInitialize()
{
    XawInitializeWidgetSet();
    XtAddConverter( XtRString, XtRJustify, XmuCvtStringToJustify, 
		    (XtConvertArgList)NULL, 0 );
}

#define TXT16 XChar2b

/*
 * Calculate width and height of displayed text in pixels
 */

static void SetTextWidthAndHeight(lw)
    JohabLabelWidget lw;
{
    XFontStruct	*fs = lw->johabLabel.font;
    XFontStruct *jfs = lw->johabLabel.johabFont; /* johab font struct */

    char *nl;

    lw->johabLabel.label_height = fs->max_bounds.ascent + fs->max_bounds.descent;
    if (lw->johabLabel.label == NULL) {
	lw->johabLabel.label_len = 0;
	lw->johabLabel.label_width = 0;
    }
    else if ((nl = index(lw->johabLabel.label, '\n')) != NULL) {
	char *label;
	lw->johabLabel.label_len = MULTI_LINE_LABEL;
	lw->johabLabel.label_width = 0;
	for (label = lw->johabLabel.label;
	     nl != NULL; nl = index(label, '\n'))
	{
	    int width;

	    width = XTextWidthJohab(fs, jfs, label, (int)(nl - label));
	    if (width > (int)lw->johabLabel.label_width)
		lw->johabLabel.label_width = width;
	    label = nl + 1;
	    if (*label)
		lw->johabLabel.label_height +=
		    fs->max_bounds.ascent + fs->max_bounds.descent;
	}
	if (*label) {
	    int width = XTextWidthJohab(fs, jfs, label, strlen(label));

	    if (width > (int) lw->johabLabel.label_width)
		lw->johabLabel.label_width = width;
	}
    }
    else {
	lw->johabLabel.label_len = strlen(lw->johabLabel.label);
	lw->johabLabel.label_width =
	    XTextWidthJohab(fs, jfs, lw->johabLabel.label,
			    (int) lw->johabLabel.label_len);
    }
}

static void GetnormalGC(lw)
    JohabLabelWidget lw;
{
    XGCValues	values;

    values.foreground	= lw->johabLabel.foreground;
    values.background	= lw->core.background_pixel;
    values.font		= lw->johabLabel.font->fid;
    values.graphics_exposures = False;

    lw->johabLabel.normal_GC = XtGetGC(
	(Widget)lw,
	(unsigned) GCForeground | GCBackground | GCFont | GCGraphicsExposures,
	&values);
}

static void GetgrayGC(lw)
    JohabLabelWidget lw;
{
    XGCValues	values;

    values.foreground = lw->johabLabel.foreground;
    values.background = lw->core.background_pixel;
    values.font	      = lw->johabLabel.font->fid;
    values.fill_style = FillTiled;
    values.tile       = XmuCreateStippledPixmap(XtScreen((Widget)lw),
						lw->johabLabel.foreground, 
						lw->core.background_pixel,
						lw->core.depth);
    values.graphics_exposures = False;

    lw->johabLabel.stipple = values.tile;

    lw->johabLabel.gray_GC = XtGetGC((Widget)lw, 
				(unsigned) GCForeground | GCBackground |
				GCFont | GCTile | GCFillStyle |
				GCGraphicsExposures,
				&values);
}

static void GetnormalJGC(lw)
    JohabLabelWidget lw;
{
    XGCValues	values;

    values.foreground	= lw->johabLabel.foreground;
    values.background	= lw->core.background_pixel;
    values.font		= lw->johabLabel.johabFont->fid;
    values.graphics_exposures = False;

    lw->johabLabel.normal_JGC = XtGetGC(
	(Widget)lw,
	(unsigned) GCForeground | GCBackground | GCFont | GCGraphicsExposures,
	&values);
}

static void GetgrayJGC(lw)
    JohabLabelWidget lw;
{
    XGCValues	values;

    values.foreground = lw->johabLabel.foreground;
    values.background = lw->core.background_pixel;
    values.font	      = lw->johabLabel.johabFont->fid;
    values.fill_style = FillTiled;
    values.tile       = XmuCreateStippledPixmap(XtScreen((Widget)lw),
						lw->johabLabel.foreground, 
						lw->core.background_pixel,
						lw->core.depth);
    values.graphics_exposures = False;

    lw->johabLabel.stipple = values.tile;

    lw->johabLabel.gray_JGC = XtGetGC((Widget)lw, 
				(unsigned) GCForeground | GCBackground |
				GCFont | GCTile | GCFillStyle |
				GCGraphicsExposures,
				 &values);
}

/* ARGSUSED */
static void Initialize(request, new, args, num_args)
    Widget request, new;
    ArgList args;
    Cardinal *num_args;
{
    JohabLabelWidget lw = (JohabLabelWidget) new;

    if (lw->johabLabel.label == NULL) 
        lw->johabLabel.label = XtNewString(lw->core.name);
    else {
        lw->johabLabel.label = XtNewString(lw->johabLabel.label);
    }

    GetnormalGC(lw);
    GetgrayGC(lw);
    GetnormalJGC(lw);
    GetgrayJGC(lw);
    
    SetTextWidthAndHeight(lw);

    if (lw->core.height == 0)
        lw->core.height = lw->johabLabel.label_height +
			    2 * lw->johabLabel.internal_height;

    if (lw->core.width == 0)
        lw->core.width = (lw->johabLabel.label_width +
			    2 * lw->johabLabel.internal_width +
			    LEFT_OFFSET(lw));

    lw->johabLabel.label_x = lw->johabLabel.label_y = 0;
    (*XtClass(new)->core_class.resize) ((Widget)lw);

} /* Initialize */

/*
 * Repaint the widget window
 */

/* ARGSUSED */
static void Redisplay(gw, event, region)
    Widget gw;
    XEvent *event;
    Region region;
{
    JohabLabelWidget w = (JohabLabelWidget) gw;
    GC gc, jgc;

    /*
     * now we'll see if we need to draw the rest of the label
     */
    if (region != NULL) {
	int x = w->johabLabel.label_x;
	unsigned int width = w->johabLabel.label_width;

	if (XRectInRegion(region, x, w->johabLabel.label_y,
			 width, w->johabLabel.label_height) == RectangleOut){
	    return;
	}
    }

    gc = XtIsSensitive(gw) ? w->johabLabel.normal_GC : w->johabLabel.gray_GC;
    jgc = XtIsSensitive(gw) ? w->johabLabel.normal_JGC : w->johabLabel.gray_JGC;
#ifdef notdef
    if (region != NULL) {
	XSetRegion(XtDisplay(gw), gc, region);
	XSetRegion(XtDisplay(gw), jgc, region);
    }
#endif /*notdef*/

    {
	int len = w->johabLabel.label_len;
	char *label = w->johabLabel.label;
	Position y = w->johabLabel.label_y +
	    w->johabLabel.font->max_bounds.ascent;

	if (len == MULTI_LINE_LABEL) {
	    char *nl;
	    while ((nl = index(label, '\n')) != NULL) {
		XDrawStringJohab(XtDisplay(gw), XtWindow(gw), gc, jgc,
				 w->johabLabel.font, w->johabLabel.johabFont,
				 w->johabLabel.label_x, y,
				 label, (int)(nl - label));
		y += w->johabLabel.font->max_bounds.ascent + 
		    w->johabLabel.font->max_bounds.descent;
		label = nl + 1;
	    }
	    len = strlen(label);
	}
	if (len) {
	    XDrawStringJohab(XtDisplay(gw), XtWindow(gw), gc, jgc,
			     w->johabLabel.font, w->johabLabel.johabFont,
			     w->johabLabel.label_x, y, label, len);
	}
    }

#ifdef notdef
    if (region != NULL) {
	XSetClipMask(XtDisplay(gw), gc, (Pixmap)None);
	XSetClipMask(XtDisplay(gw), jgc, (Pixmap)None);
    }
#endif /* notdef */
}

static void _Reposition(lw, width, height, dx, dy)
    JohabLabelWidget lw;
    Dimension width, height;
    Position *dx, *dy;
{
    Position newPos;
    Position leftedge = lw->johabLabel.internal_width + LEFT_OFFSET(lw);

    switch (lw->johabLabel.justify) {

	case XtJustifyLeft   :
	    newPos = leftedge;
	    break;

	case XtJustifyRight  :
	    newPos = width -
		(lw->johabLabel.label_width + lw->johabLabel.internal_width);
	    break;

	case XtJustifyCenter :
	default:
	    newPos = (int)(width - lw->johabLabel.label_width) / 2;
	    break;
    }
    if (newPos < (Position)leftedge)
	newPos = leftedge;
    *dx = newPos - lw->johabLabel.label_x;
    lw->johabLabel.label_x = newPos;
    *dy = (newPos = (int)(height - lw->johabLabel.label_height) / 2)
	  - lw->johabLabel.label_y;
    lw->johabLabel.label_y = newPos;
    return;
}

static void Resize(w)
    Widget w;
{
    JohabLabelWidget lw = (JohabLabelWidget)w;
    Position dx, dy;

    _Reposition(lw, w->core.width, w->core.height, &dx, &dy);
}

/*
 * Set specified arguments into widget
 */

#define PIXMAP 0
#define WIDTH 1
#define HEIGHT 2
#define NUM_CHECKS 3

static Boolean SetValues(current, request, new, args, num_args)
    Widget current, request, new;
    ArgList args;
    Cardinal *num_args;
{
    JohabLabelWidget curlw = (JohabLabelWidget) current;
    JohabLabelWidget reqlw = (JohabLabelWidget) request;
    JohabLabelWidget newlw = (JohabLabelWidget) new;
    int i;
    Boolean was_resized = False, redisplay = False, checks[NUM_CHECKS];

    for (i = 0; i < NUM_CHECKS; i++)
	checks[i] = FALSE;

    for (i = 0; i < *num_args; i++) {
	if (streq(XtNwidth, args[i].name))
	    checks[WIDTH] = TRUE;
	if (streq(XtNheight, args[i].name))
	    checks[HEIGHT] = TRUE;
    }

    if (newlw->johabLabel.label == NULL) {
	newlw->johabLabel.label = newlw->core.name;
    }

    if (curlw->johabLabel.label != newlw->johabLabel.label) {
        if (curlw->johabLabel.label != curlw->core.name)
	    XtFree( (char *)curlw->johabLabel.label );

	if (newlw->johabLabel.label != newlw->core.name) {
	    newlw->johabLabel.label = XtNewString( newlw->johabLabel.label );
	}
	was_resized = True;
    }

    if (was_resized || (curlw->johabLabel.font != newlw->johabLabel.font) ||
	(curlw->johabLabel.johabFont != newlw->johabLabel.johabFont) ||
	(curlw->johabLabel.justify != newlw->johabLabel.justify) || checks[PIXMAP])
    {
	SetTextWidthAndHeight(newlw);
	was_resized = True;
    }

    /* recalculate the window size if something has changed. */
    if (newlw->johabLabel.resize && was_resized) {
	if ((curlw->core.height == reqlw->core.height) && !checks[HEIGHT])
	    newlw->core.height = (newlw->johabLabel.label_height +
				    2 * newlw->johabLabel.internal_height);

	if ((curlw->core.width == reqlw->core.width) && !checks[WIDTH])
	    newlw->core.width = (newlw->johabLabel.label_width +
				    LEFT_OFFSET(newlw) +
				    2 * newlw->johabLabel.internal_width);
    }

    if (curlw->johabLabel.foreground        != newlw->johabLabel.foreground
	|| curlw->core.background_pixel     != newlw->core.background_pixel
	|| curlw->johabLabel.font->fid      != newlw->johabLabel.font->fid
	|| curlw->johabLabel.johabFont->fid !=
	newlw->johabLabel.johabFont->fid)
    {
        /* The Fontset is not in the GC - don't make a new GC if FS changes! */

	XtReleaseGC(new, curlw->johabLabel.normal_GC);
	XtReleaseGC(new, curlw->johabLabel.gray_GC);
	XtReleaseGC(new, curlw->johabLabel.normal_JGC);
	XtReleaseGC(new, curlw->johabLabel.gray_JGC);
	XmuReleaseStippledPixmap( XtScreen(current), curlw->johabLabel.stipple );
	GetnormalGC(newlw);
	GetgrayGC(newlw);
	GetnormalJGC(newlw);
	GetgrayJGC(newlw);
	redisplay = True;
    }

    if ((curlw->johabLabel.internal_width != newlw->johabLabel.internal_width)
        || (curlw->johabLabel.internal_height != newlw->johabLabel.internal_height)
	|| was_resized) {
	/* Resize() will be called if geometry changes succeed */
	Position dx, dy;
	_Reposition(newlw, curlw->core.width, curlw->core.height, &dx, &dy);
    }

    return was_resized || redisplay ||
	   XtIsSensitive(current) != XtIsSensitive(new);
}

static void Destroy(w)
    Widget w;
{
    JohabLabelWidget lw = (JohabLabelWidget)w;

    if ( lw->johabLabel.label != lw->core.name )
	XtFree( lw->johabLabel.label );
    XtReleaseGC( w, lw->johabLabel.normal_GC );
    XtReleaseGC( w, lw->johabLabel.gray_GC);
    XtReleaseGC( w, lw->johabLabel.normal_JGC );
    XtReleaseGC( w, lw->johabLabel.gray_JGC);
    XmuReleaseStippledPixmap( XtScreen(w), lw->johabLabel.stipple );
}


static XtGeometryResult QueryGeometry(w, intended, preferred)
    Widget w;
    XtWidgetGeometry *intended, *preferred;
{
    JohabLabelWidget lw = (JohabLabelWidget)w;

    preferred->request_mode = CWWidth | CWHeight;
    preferred->width = (lw->johabLabel.label_width + 
			    2 * lw->johabLabel.internal_width +
			    LEFT_OFFSET(lw));
    preferred->height = lw->johabLabel.label_height + 
			    2 * lw->johabLabel.internal_height;
    if (  ((intended->request_mode & (CWWidth | CWHeight))
	   	== (CWWidth | CWHeight)) &&
	  intended->width == preferred->width &&
	  intended->height == preferred->height)
	return XtGeometryYes;
    else if (preferred->width == w->core.width &&
	     preferred->height == w->core.height)
	return XtGeometryNo;
    else
	return XtGeometryAlmost;
}
