// -*- mode: objc; tab-width: 4 -*-

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#import  "KAutomataForX.h"
#include "sutil.h"

#define IS_HAN(jamokey) ((jamokey) >= XK_Hangul_Kiyeog && \
						 (jamokey) <= XK_Hangul_J_Hieuh)

typedef int (*QsortComp)(const void *, const void *);
typedef QsortComp BsearchComp;

// Methods modifying these variables must not recurse.
static char *current_file;
static Display *current_display;

static struct keysym_to_keycode {
	KeySym keysym;
	KeyCode keycode;
} *keysym_to_keycode_table;

static XModifierKeymap *mod_map;

static int count_lines(char *p, char *eof);
static BOOL prepare_buf(char **buf, int *buf_size, int required);

@implementation KAutomataForX

- init
{
	// don't call me!
	[self free];
	return nil;
}

- initFromFile:(char *)path on:(Display *)display
{
	int fd;
	char *file;
	int file_size;
	
	if ((self = [super init])) {
		if ((fd = open(path, O_RDONLY)) >= 0) {
			file_size = lseek(fd, 0, SEEK_END);
			if ((file = mmap(NULL, file_size, PROT_READ, MAP_SHARED, fd, 0)) !=
				(void *)-1)
			{
				id result;
			
				current_file = path;
				current_display = display;
				result = [self readKDF:file whoseSize:file_size];
				munmap(file, file_size);
				close(fd);
				if (result)
					return self;
				else
					goto fail;
			}
			perror("KAutomataForX");
			close(fd);
			goto fail;
		}
		perror("KAutomataForX");
	}
fail:
	[self free];
	return nil;
}

- free
{
	if (map != NULL)
		free(map);

	[super free];

	return nil;
}

- (int)eatKeycode:(KeyCode)keycode modifier:(unsigned int)modifier
{
	struct keycode_to_jamokey_map *e;
	struct keycode_to_jamokey_map t = { keycode };
	static int compare_keycodes(struct keycode_to_jamokey_map *e1,
								struct keycode_to_jamokey_map *e2);
	
	if ((e = bsearch(&t, map, njamokeys, sizeof(map[0]),
					 (BsearchComp)compare_keycodes)) == NULL)
		return WRONGKEY;		// don't know what the key is

	for (; e > map && (e-1)->keycode == t.keycode; e--)
		;
	for (; e < map + njamokeys && e->keycode == t.keycode; e++)
		if ((modifier & e->modifier_mask) == e->modifier) {
			int r;
			
			if (IS_HAN(e->candidates[0].jamo_key)) {
				if ((r = [super eatKey:e->candidates]) == CHARBREAK)
					strcpy(composed_chars, [super getComposedChar]);
				return r;
			}
			else {
				sprintf(composed_chars, "%s%c", [super getComposingChar],
						e->candidates[0].jamo_key);
				[super reset];
				return CHARBREAK;
			}
		}
	
	return WRONGKEY;
}

- (char *)getComposedChar
{
	return composed_chars;
}

- readKDF:(char *)f whoseSize:(int)size
{
	char *eof = f + size;
	char *p;
	int l;
	char *buf = NULL;
	int buf_size = 0;
	char *implicit_keys = f, *implicit_keys_end = f;
	int num_non_comment_lines = 0;
	static int compare_map_entries(struct keycode_to_jamokey_map *e1,
								   struct keycode_to_jamokey_map *e2);
	
	// num_non_comment_lines is greater than number of key definitions
	// by the number of `Implicit:' lines.  But that doesn' matter very much.
	num_non_comment_lines = count_lines(f, eof);
	if ((map = malloc(num_non_comment_lines * sizeof(map[0]))) == NULL) {
		perror("KAutomataForX");
		goto fail;
	}

	// parse key definitions.

	for (p = f, l = 1;
		 p = eat_space(p, eof), p < eof;
		 p = next_line(p, eof), l++)
	{
		char *eol;
		char *colon;
		char *q;
		KeySym keysym;
		KeyCode keycode;
		unsigned int modifier, modifier_mask;
		JamoKey jamokey;
		JamoKeyAndStateOffset candidates[MAXCANDIDATES];
		int i;
		
		if (*p == '!' || *p == '\n') // comment or blank line
			continue;

		eol = next_occurrence_of('\n', p, eof);
		if ((colon = next_occurrence_of(':', p, eol)) == eol) {	// no colon
			fprintf(stderr, "%s:%d: colon(`:') not found.\n", current_file, l);
			goto fail;
		}

		// treat implicit directives.

		if (memstrcmp(p, previous_occurrence_of_nonspace(colon, p) - p,
					  "Implicit") == 0) // `Implicit:' directive
		{
			implicit_keys = colon + 1;
			implicit_keys_end = eol;
			continue;
		}
		
		// parse keycode and modifiers (left part of `:')
		
		if (!prepare_buf(&buf, &buf_size, (colon-p) +
						 (implicit_keys_end-implicit_keys) + 2))
			goto fail;
		memcpy(buf, p, colon - p);
		buf[colon - p] = ' ';
		strmemcpy(buf + (colon-p) + 1, implicit_keys,
				  implicit_keys_end - implicit_keys);

		if (!parse_key(&keysym, &modifier, &modifier_mask, buf)) {
			fprintf(stderr, "%s:%d: parse error.\n", current_file, l);
			goto fail;
		}
		if ((keycode = XKeysymToKeycode(current_display, keysym)) == 0) {
			fprintf(stderr, "%s:%d: cannot convert keysym `%s' to keycode.\n",
					current_file, l, XKeysymToString(keysym));
			goto fail;
		}
		
		// parse what to produce

		for (i = 0, p = colon + 1;
			 i < MAXCANDIDATES && (p = eat_space(p, eol), p != eol);
			 p = q)
		{
			q = end_of_word(p, eol);
			if (!prepare_buf(&buf, &buf_size, q - p + 1))
				goto fail;
			strmemcpy(buf, p, q - p);

			if ((jamokey = XStringToKeysym(buf)) == NoSymbol) {
				fprintf(stderr, "%s:%d: strange keysym `%s'.\n",
						current_file, l, buf);
				goto fail;
			}
			if (strncmp(buf, "Hangul_", 7) == 0) { // Korean jamo
				char *r;
				
				p = eat_space(q, eol);
				q = end_of_word(p, eol);
				
				for (r = p; r < p+MAXCANDIDATES && r < q; r++) {
					if (strchr("12345", *r) == NULL) {
						fprintf(stderr, "%s:%d: `%c' is not one of 1-5.\n",
								current_file, l, *r);
						goto fail;
					}
					candidates[i].jamo_key = jamokey;
					candidates[i].state_offset = *r - '1';
					i++;
				}
			}
			else {
				candidates[i].jamo_key = jamokey;
				i++;
			}
		} // for

		if (i < MAXCANDIDATES)
			candidates[i].jamo_key = 0;
		
		map[njamokeys].keycode = keycode;
		map[njamokeys].modifier = modifier;
		map[njamokeys].modifier_mask = modifier_mask;
		memcpy(map[njamokeys].candidates, candidates, sizeof(candidates));
		njamokeys++;
	} // for

	map = realloc(map, njamokeys * sizeof(map[0]));
	qsort(map, njamokeys, sizeof(map[0]), (QsortComp)compare_map_entries);

	if (buf != NULL)
		free(buf);
	return self;

fail:
	if (map != NULL) {
		free(map);
		map = NULL;
	}
	if (buf != NULL)
		free(buf);
	return nil;
}

@end

static int compare_map_entries(struct keycode_to_jamokey_map *e1,
							   struct keycode_to_jamokey_map *e2)
{
	if (e1->keycode != e2->keycode)
		return e1->keycode - e2->keycode;
	if (e1->modifier != e2->modifier)
		return (signed int)e1->modifier - (signed int)e2->modifier;
	if (e1->modifier_mask != e2->modifier_mask)
		return (signed int)e1->modifier_mask - (signed int)e2->modifier_mask;
	return 0;
}

static int compare_keycodes(struct keycode_to_jamokey_map *e1,
							struct keycode_to_jamokey_map *e2)
{
	return (signed int)e1->keycode - (signed int)e2->keycode;
}

static int count_lines(char *p, char *eof)
{
	int n = 0;
	
	for (; p = eat_space(p, eof), p < eof; p = next_line(p, eof))
		if (*p != '!' || *p != '\n')
			n++;

	return n;
}

static BOOL prepare_buf(char **buf, int *buf_size, int required)
{
	if (buf == NULL || *buf_size < required) {
		char *new_buf;
			
		if ((new_buf = realloc(*buf, required)) == NULL) {
			perror("KAutomataForX");
			return NO;
		}

		*buf = new_buf;
		*buf_size = required;
	}

	return YES;
}

BOOL is_modifier(KeyCode keycode)
{
	int n;
	KeyCode *p;
	
	for (p = mod_map->modifiermap, n = 8 * mod_map->max_keypermod;
		 n > 0; n--, p++)
		if (keycode != None && *p == keycode)
			return YES;

	return NO;
}

BOOL retrieve_keyboard_mapping(Display *display)
{
	int min_keycode, max_keycode;
	KeySym *keysym_list;
	int keysyms_per_keycode;
	int i, j, n;
	static int compare_keysyms(struct keysym_to_keycode *e1,
							   struct keysym_to_keycode *e2);
	
	if ((mod_map = XGetModifierMapping(display)) == NULL) {
		fprintf(stderr,
				"retrieve_keyboard_mapping(): Cannot get modifier mapping.\n");
		return NO;
	}
	
	XDisplayKeycodes(display, &min_keycode, &max_keycode);
	n = max_keycode - min_keycode + 1;

	if (!(keysym_list = XGetKeyboardMapping(display, min_keycode, n,
											&keysyms_per_keycode)))
	{
		fprintf(stderr,
				"retrieve_keyboard_mapping(): Cannot get keyboard mapping.\n");
		return NO;
	}
	if ((keysym_to_keycode_table = malloc(n * keysyms_per_keycode *
		 sizeof(keysym_to_keycode_table[0]))) == NULL)
	{
		perror("retrieve_keyboard_mapping()");
		XFree(keysym_list);
		return NO;
	}

	for (i = 0; min_keycode <= max_keycode; min_keycode++)
		for (j = keysyms_per_keycode; j > 0; j--, i++) {
			keysym_to_keycode_table[i].keysym = keysym_list[i];
			keysym_to_keycode_table[i].keycode = min_keycode;
		}
	XFree(keysym_list);
	qsort(keysym_to_keycode_table, n, sizeof(keysym_to_keycode_table[0]),
		  (QsortComp)compare_keysyms);

	return YES;
}

static int compare_keysyms(struct keysym_to_keycode *e1,
						   struct keysym_to_keycode *e2)
{
	return e1->keysym - e2->keysym;
}
