// -*- mode: objc; tab-width: 4 -*-

#include "Intrinsic.h"
#import  <objc/Object.h>

@class KAutomataForX;

struct trigger_key {
	KeySym keysym;
	unsigned int modifier;
	unsigned int modifier_mask;
};

typedef enum { MRAW, MTRANS, MKOREAN } KeyboardType;

@interface KText : Object
{
	KeyboardType keyboard;
	KAutomataForX *am;
	BOOL phantom;				// indicates that before the caret is a
								// phantom character.
	Widget text;
	EventMask text_event_mask;
	Widget translator;			// Widget on which text translations are
								// installed.
	struct trigger_key toggle;
	struct trigger_key conversion;
}

- init;							// this must not be called.
- initFromKDF:(char *)kdf_path withWidget:(Widget)text translator:(Widget)tw;

- setKTT:(char *)path;
- setKDF:(char *)path;
- setToggleKey:(struct trigger_key *)key;
- setConversionKey:(struct trigger_key *)key;
- setKeyboard:(KeyboardType)keyboard;
- (Widget)getWidget;
- setString:(char *)string;
- (char *)getString;

//
//  private methods
//

// Text widget action methods

- (void)textForwardCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textForwardWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textForwardParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBeginningOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textEndOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNextLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textPreviousLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNextPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textPreviousPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBeginningOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textEndOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textScrollOneLineUp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textScrollOneLineDown:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeleteNextCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeletePreviousCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeleteNextWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeletePreviousWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDeleteSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textBackwardKillWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillToEndOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textKillToEndOfParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNewlineAndIndent:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNewlineAndBackup:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNewline:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectAll:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectStart:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectAdjust:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectEnd:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSelectSave:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textExtendStart:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textExtendAdjust:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textExtendEnd:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textRedrawDisplay:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSearch:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertChar:(Widget)w :(XKeyEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertString:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textFocusIn:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textFocusOut:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textEnterWindow:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textLeaveWindow:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDisplayCaret:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textMultiply:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textFormParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textTransposeCharacters:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textNoOp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textInsertFileAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDoSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textDoReplaceAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textSetField:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textPopdownSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;
- (void)textReconnectIm:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np;

- deletePhantomIfPresent;
- insertPhantom;
- insertComposingChar;

@end

// This function must not be called.
void _KTextEventHandler(Widget w, KText *self, XEvent *ev, Boolean *cont);
