// -*- mode: objc; tab-width: 4 -*-

#include <stdio.h>
#include "Intrinsic.h"
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>

#import  "KAutomataForX.h"
#import  "KText.h"
#import  "textActions.h"

#include "hangul.h"

#define MAXCODECONVERTEDTEXTLEN 20

static void insert_string_in_text(Widget w, char *string_in_wansung);
static BOOL is_trigger(XKeyEvent *kev, struct trigger_key *tr);

@implementation KText

- init
{
	[self free];
	return nil;
}

- initFromKDF:(char *)kdf_path withWidget:(Widget)aText translator:(Widget)tw
{
	text = aText;
	
	if ([self = [super init] setKDF:kdf_path]) {
		Widget dummyText;
		XtTranslations text_translations;

		if ((dummyText = XtVaCreateManagedWidget(XtName(text),
										     asciiTextWidgetClass, tw, NULL)))
		{
			XtVaGetValues(dummyText,
						  XtNtranslations, &text_translations, NULL);
			XtOverrideTranslations(tw, text_translations);
			XtDestroyWidget(dummyText);
			translator = tw;

			text_event_mask = XtBuildEventMask(text);
			XtUninstallTranslations(text);
			XtAddEventHandler(text,
							  text_event_mask, False,
							  (XtEventHandler)_KTextEventHandler, self);

			toggle.keysym = XK_space;
			toggle.modifier = ShiftMask;
			toggle.modifier_mask = ShiftMask | ControlMask |
				Mod1Mask | Mod2Mask | Mod3Mask | Mod4Mask | Mod5Mask;
			conversion.keysym = XK_F9;
			conversion.modifier = Mod1Mask;
			conversion.modifier_mask = toggle.modifier_mask;
			
			return self;
		}
	}

	[self free];
	return nil;
}

- free
{
	XtDestroyWidget(text);
	[am free];
	[super free];
	return nil;
}

- setKDF:(char *)kdf_path
{
	KAutomataForX *new_am;

	if (!(new_am = [[KAutomataForX alloc] initFromFile:kdf_path
										  on:XtDisplay(text)]))
		return nil;
	
	[am free];
	am = new_am;
	return self;
}

- setToggleKey:(struct trigger_key *)key
{
	toggle = *key;
	return self;
}
- setConversionKey:(struct trigger_key *)key
{
	conversion = *key;
	return self;
}

- setKeyboard:(KeyboardType)new_keyboard
{
	if (keyboard == MKOREAN)
		[[self deletePhantomIfPresent] insertComposingChar];
	keyboard = new_keyboard;
	return self;
}

- (Widget)getWidget
{
	return text;
}

- setString:(char *)string
{
	XtVaSetValues(text, XtNstring, string, NULL);
	return self;
}

- (char *)getString
{
	char *string;

	[[self deletePhantomIfPresent] insertComposingChar];
	XtVaGetValues(text, XtNstring, &string, NULL);
	return string;
}

//
//  private methods
//

//
//  Text widget action methods
//

- (void)textForwardCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "forward-character", ev, p, *np);
}

- (void)textBackwardCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	if (phantom) {
		XawTextPosition p;

		[self deletePhantomIfPresent];
		p = XawTextGetInsertionPoint(text);
		[self insertComposingChar];
		XawTextSetInsertionPoint(text, p);
	}
	else
		XtCallActionProc(text, "backward-character", ev, p,
						 *np);
}

- (void)textForwardWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "forward-word", ev, p, *np);
}

- (void)textBackwardWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "backward-word", ev, p, *np);
}

- (void)textForwardParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "forward-paragraph", ev, p, *np);
}

- (void)textBackwardParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "backward-paragraph", ev, p, *np);
}

- (void)textBeginningOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "beginning-of-line", ev, p, *np);
}

- (void)textEndOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "end-of-line", ev, p, *np);
}

- (void)textNextLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "next-line", ev, p, *np);
}

- (void)textPreviousLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "previous-line", ev, p, *np);
}

- (void)textNextPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "next-page", ev, p, *np);
}

- (void)textPreviousPage:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "previous-page", ev, p, *np);
}

- (void)textBeginningOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "beginning-of-file", ev, p, *np);
}

- (void)textEndOfFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "end-of-file", ev, p, *np);
}

- (void)textScrollOneLineUp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "end-of-file", ev, p, *np);
}

- (void)textScrollOneLineDown:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "scroll-one-line-down", ev, p, *np);
}

- (void)textDeleteNextCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "delete-next-character", ev, p, *np);
}

- (void)textDeletePreviousCharacter:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[self deletePhantomIfPresent];
	if ([am deleteLastKey] == EMPTY)
		XtCallActionProc(text, "delete-previous-character",
						 ev, p, *np);
	else
		[self insertPhantom];
}

- (void)textDeleteNextWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "delete-next-word", ev, p, *np);
}

- (void)textDeletePreviousWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "delete-previous-word", ev, p, *np);
}

- (void)textDeleteSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "delete-selection", ev, p, *np);
}

- (void)textKillWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "kill-word", ev, p, *np);
}

- (void)textBackwardKillWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "backward-kill-word", ev, p, *np);
}

- (void)textKillSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "kill-selection", ev, p, *np);
}

- (void)textKillToEndOfLine:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "kill-to-end-of-line", ev, p, *np);
}

- (void)textKillToEndOfParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "kill-to-end-of-paragraph", ev, p, *np);
}

- (void)textNewlineAndIndent:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "newline-and-indent", ev, p, *np);
}

- (void)textNewlineAndBackup:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "newline-and-backup", ev, p, *np);
}

- (void)textNewline:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "newline", ev, p, *np);
}

- (void)textSelectWord:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "select-word", ev, p, *np);
}

- (void)textSelectAll:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "select-all", ev, p, *np);
}

- (void)textSelectStart:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "select-start", ev, p, *np);
}

- (void)textSelectAdjust:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "select-adjust", ev, p, *np);
}

- (void)textSelectEnd:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "select-end", ev, p, *np);
}

- (void)textSelectSave:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "select-save", ev, p, *np);
}

- (void)textExtendStart:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "extend-start", ev, p, *np);
}

- (void)textExtendAdjust:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "extend-adjust", ev, p, *np);
}

- (void)textExtendEnd:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "extend-end", ev, p, *np);
}

- (void)textInsertSelection:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "insert-selection", ev, p, *np);
}

- (void)textRedrawDisplay:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "redraw-display", ev, p, *np);
}

- (void)textInsertFile:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "insert-file", ev, p, *np);
}

- (void)textSearch:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "search", ev, p, *np);
}

- (void)textInsertChar:(Widget)w :(XKeyEvent *)kev :(String *)p :(Cardinal *)np
{
	if (kev->type == KeyPress) {
		if (is_trigger(kev, &toggle))
			if (keyboard == MKOREAN)
				[self setKeyboard:MRAW];
			else
				[self setKeyboard:MKOREAN];
		else if (is_trigger(kev, &conversion)) {
		}
		else if (!is_modifier(kev->keycode))
			switch (keyboard) {
				char wansung_text[MAXCODECONVERTEDTEXTLEN];

			case MKOREAN:
				[self deletePhantomIfPresent];
				switch ([am eatKeycode:kev->keycode modifier:kev->state]) {
				case CHARBREAK:
					wansung_text[convert_johab_to_ks([am getComposedChar],
						 wansung_text, strlen([am getComposedChar]))] = '\0';
					insert_string_in_text(text, wansung_text);
				case JUSTFINE:
					[self insertPhantom];
					break;
				case WRONGKEY:
					[self insertComposingChar];
					XtCallActionProc(text,
									 "insert-char", (XEvent *)kev, p, *np);
					break;
				default:
					fprintf(stderr, "automata error.\n");
				}
				break;
			case MRAW:
				XtCallActionProc(text, "insert-char", (XEvent *)kev, p, *np);
				break;
			case MTRANS:
				break;
			}
		else					// Let Text widget handle the event
			XtCallActionProc(text, "insert-char", (XEvent *)kev, p, *np);
	} // KeyPress
	else						// other than KeyPress
		XtCallActionProc(text, "insert-char", (XEvent *)kev, p, *np);
}

- (void)textInsertString:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "insert-string", ev, p, *np);
}

- (void)textFocusIn:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "focus-in", ev, p, *np);
}

- (void)textFocusOut:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "focus-out", ev, p, *np);
}

- (void)textEnterWindow:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "enter-window", ev, p, *np);
}

- (void)textLeaveWindow:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "leave-window", ev, p, *np);
}

- (void)textDisplayCaret:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "display-caret", ev, p, *np);
}

- (void)textMultiply:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "multiply", ev, p, *np);
}

- (void)textFormParagraph:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "form-paragraph", ev, p, *np);
}

- (void)textTransposeCharacters:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "transpose-characters", ev, p, *np);
}

- (void)textNoOp:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "no-op", ev, p, *np);
}

- (void)textInsertFileAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "insert-file-action", ev, p, *np);
}

- (void)textDoSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "do-search-action", ev, p, *np);
}

- (void)textDoReplaceAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "do-replace-action", ev, p, *np);
}

- (void)textSetField:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	[[self deletePhantomIfPresent] insertComposingChar];
	XtCallActionProc(text, "set-field", ev, p, *np);
}

- (void)textPopdownSearchAction:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	XtCallActionProc(text, "popdown-search-action", ev, p, *np);
}

- (void)textReconnectIm:(Widget)w :(XEvent *)ev :(String *)p :(Cardinal *)np
{
	// never attempt to connect to myself!
}

- deletePhantomIfPresent
{
	if (phantom) {
		XtCallActionProc(text,
						 "delete-previous-character", NULL, NULL, 0);
		phantom = NO;
	}

	return self;
}

- insertPhantom
{
	if (strcmp([am getComposingChar], "") != 0) {
		char wansung_text[MAXCODECONVERTEDTEXTLEN];

		wansung_text[convert_johab_to_ks([am getComposingChar], wansung_text,
										strlen([am getComposingChar]))] = '\0';
		wansung_text[2] = '\0';
		insert_string_in_text(text, wansung_text);
		phantom = YES;
	}

	return self;
}

- insertComposingChar
{
	if ([am getComposingChar][0]) {
		char wansung_text[MAXCODECONVERTEDTEXTLEN];

		wansung_text[convert_johab_to_ks([am getComposingChar], wansung_text,
									 strlen([am getComposingChar]))] = '\0';
		insert_string_in_text(text, wansung_text);
		[am reset];
	}
	return self;
}

@end

void _KTextEventHandler(Widget w, KText *self, XEvent *ev,
						Boolean *continue_to_dispatch)
{
	struct KTextStruct { @defs(KText) } *ktext;
	
	ktext = (struct KTextStruct *)(_eReceiver = self);
	XtDispatchEventToWidget(ktext->translator, ev);
}

static void insert_string_in_text(Widget w, char *string_in_wansung)
{
	XtCallActionProc(w, "insert-string", NULL, &string_in_wansung, 1);
}

static BOOL is_trigger(XKeyEvent *kev, struct trigger_key *tr)
{
	if (XLookupKeysym(kev, 0) == tr->keysym &&
		(kev->state & tr->modifier_mask) == tr->modifier)
		return YES;
	else
		return NO;
}
