// -*- mode: objc; tab-width: 4 -*-

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include "Intrinsic.h"
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include "JohabLabel.h"

#import  "ICManager.h"
#import  "IContext.h"
#import  "KAutomataForX.h"
#include "textActions.h"

#include "byeoroo.h"
#include "sutil.h"

#define XtNtransport "transport"
#define XtCTransport "Transport"
#define XtNkorEngToggleKey "korEngToggleKey"
#define XtCKorEngToggleKey "KorEngToggleKey"
#define XtNchineseConversionKey "chineseConversionKey"
#define XtCChineseConversionKey "ChineseConversionKey"
#define XtNpopupKey "popupKey"
#define XtCPopupKey "PopupKey"
#define XtNsearchPath "searchPath"
#define XtCSearchPath "SearchPath"
#define XtNdefaultKeyboardDescriptionFile "defaultKeyboardDescriptionFile"
#define XtCDefaultKeyboardDescriptionFile "DefaultKeyboardDescriptionFile"

#define off(f) XtOffsetOf(struct app_data, f)
static XtResource resources[] = {
	{
		XtNtransport,
		XtCTransport,
		XtRString,
		sizeof(char *),
		off(transport),
		XtRImmediate,
		NULL
	},
	{
		XtNkorEngToggleKey,
		XtCKorEngToggleKey,
		XtRString,
		sizeof(char *),
		off(toggle_key),
		XtRImmediate,
		NULL
	},
	{
		XtNchineseConversionKey,
		XtCChineseConversionKey,
		XtRString,
		sizeof(char *),
		off(chinese_key),
		XtRImmediate,
		NULL
	},
	{
		XtNpopupKey,
		XtCPopupKey,
		XtRString,
		sizeof(char *),
		off(popup_key),
		XtRImmediate,
		NULL
	},
	{
		XtNsearchPath,
		XtCSearchPath,
		XtRString,
		sizeof(char *),
		off(search_path),
		XtRImmediate,
		NULL
	},
	{
		XtNdefaultKeyboardDescriptionFile,
		XtCDefaultKeyboardDescriptionFile,
		XtRString,
		sizeof(char *),
		off(default_kdf),
		XtRImmediate,
		NULL
	}
};
#undef off

static XrmOptionDescRec options[] = {
	{ "-transport", "*" XtNtransport, XrmoptionSepArg, NULL },
	{ "-kdf", "*" XtNdefaultKeyboardDescriptionFile, XrmoptionSepArg, NULL },
};

struct app_data app_data;
XtAppContext app_context;
XIMS ims;						// input method service
Window iwin;					// Window for transport
Display *display;				
XOM om;							// output method
IContext *focusedIC = nil;		// currently focused IC
IContext *lastFocusedIC = nil;
BOOL cp_popped_up = NO;

//
//  main window widgets
//

Widget topLevel;
Widget box;
Widget korEngStat;
Widget charLineStat;
Widget kdfLabel;
Widget kttLabel;

//
//  control panel widgets
//

Widget cpshell;
Widget cpform;
Widget cptitle;
Widget cplabel1;
Widget cplabel2;
Widget cplabel3;
Widget cpchar;
Widget cpline;
Widget cpkdf;
Widget cpktt;
Widget cpokay;
Widget cpcancel;
Widget cplistshell;
Widget cplist;

static XIMTriggerKey trigger_keys[] = {
	{ 0L, 0L, 0L },				// Korean-English key
	{ 0L, 0L, 0L },				// Popup key
	{ 0L, 0L, 0L }				// Chinese conversion key
};
static XIMTriggerKeys trigger_key_list = {
	XtNumber(trigger_keys),
	trigger_keys
};

static XIMStyle styles[] = {
    XIMPreeditCallbacks | XIMStatusCallbacks,
    XIMPreeditCallbacks | XIMStatusArea,     
    XIMPreeditCallbacks | XIMStatusNothing,  
    XIMPreeditCallbacks | XIMStatusNone,     
    XIMPreeditPosition  | XIMStatusCallbacks,
    XIMPreeditPosition  | XIMStatusArea,     
    XIMPreeditPosition  | XIMStatusNothing,  
    XIMPreeditPosition  | XIMStatusNone,     
    XIMPreeditArea      | XIMStatusCallbacks,
    XIMPreeditArea      | XIMStatusArea,     
    XIMPreeditArea      | XIMStatusNothing,  
    XIMPreeditArea      | XIMStatusNone,     
    XIMPreeditNothing   | XIMStatusCallbacks,
    XIMPreeditNothing   | XIMStatusArea,     
    XIMPreeditNothing   | XIMStatusNothing,  
    XIMPreeditNothing   | XIMStatusNone,     
    XIMPreeditNone      | XIMStatusCallbacks,
    XIMPreeditNone      | XIMStatusArea,     
    XIMPreeditNone      | XIMStatusNothing,  
    XIMPreeditNone      | XIMStatusNone      
};
static XIMStyles style_list = { XtNumber(styles), styles };

static XIMEncoding encodings[] = {
	"COMPOUND_TEXT"				// This one suffices.
};
static XIMEncodings encoding_list = { XtNumber(encodings), encodings };

static char *program_name;		// used when printing an error message.
static ICManager *theICManager;

// These two variables are referenced only in CPPopupFileList() directly
// and in CPFileChosen() indirectly. (via call_data->string)
static char **file_list = NULL;	// used to construct a file name list widget.
static int file_list_size = 0;
static int nfiles;				// number of file names

static void create_widgets();
static void open_om();
static void parse_resources();
static void help_msg(char *argv[]);

static int protocol_handler(XIMS ims, IMProtocol *call_data);
static int open_handler(XIMS, IMOpenStruct *);
static int close_handler(XIMS, IMCloseStruct *);
static int create_ic_handler(XIMS, IMChangeICStruct *);
static int destroy_ic_handler(XIMS, IMDestroyICStruct *);

static void CPPopupFileList(Widget w, char *cl_d, XtPointer c_d);
static void CPFileChosen(Widget w, XtPointer cl_d, XawListReturnStruct *c_d);
static void CPOkay(Widget w, XtPointer client_data, XtPointer call_data);
static void CPCancel(Widget w, XtPointer client_data, XtPointer call_data);

int main(int argc, char *argv[])
{
	program_name = argv[0];

	if ((theICManager = [[ICManager alloc] init]) == nil)
		print_e(1, "Cannot manage input contexts.\n");
	
	XtSetLanguageProc(NULL, NULL, NULL);
	topLevel = XtVaOpenApplication(&app_context, APPCLASS,
								   options, XtNumber(options),
								   &argc, argv,
								   NULL, // fallback resources
								   sessionShellWidgetClass,
								   NULL);
	if (argc > 1) {
		print_e(0, "Bad command line option: %s\n", argv[1]);
		help_msg(argv);
		exit(1);
	}
	XtVaGetApplicationResources(topLevel,
								&app_data,
								resources, XtNumber(resources),
								NULL);
	parse_resources();

	create_widgets();
	XtRealizeWidget(topLevel);
	install_text_actions(app_context);
	
	display = XtDisplay(topLevel);
	iwin = XtWindow(topLevel);
	open_om();
	if (!retrieve_keyboard_mapping(display))
		exit(1);
	
	if ((ims = IMOpenIM(display,
						IMModifiers, "Xi18n", // We use R6 XIM protocol.
						IMServerWindow, iwin,
						IMServerName, IMSNAME,
						IMLocale, IMSLOCALE,
						IMServerTransport, app_data.transport,
						IMInputStyles, &style_list,
						IMEncodingList, &encoding_list,
						IMFilterEventMask, FILTEREVENTMASK,
						IMOnKeysList, &trigger_key_list,
						IMProtocolHandler, protocol_handler,
						NULL)) == NULL)
		print_e(1, "Cannot open input method service.\n");
	
	XtAppMainLoop(app_context);

	return -5;			  // if control reaches here, something must be wrong.
}

static void create_widgets()
{
	XFontStruct *jf;
	Dimension iw, ih;
	
	box = XtVaCreateManagedWidget("box", boxWidgetClass, topLevel, NULL);
	korEngStat = XtVaCreateManagedWidget("korEngStat", johabLabelWidgetClass,
										 box, NULL);
	charLineStat = XtVaCreateManagedWidget("charLineStat", labelWidgetClass,
										   box, NULL);
	kdfLabel = XtVaCreateManagedWidget("kdfLabel", labelWidgetClass, box,
									   NULL);
	kttLabel = XtVaCreateManagedWidget("kttLabel", labelWidgetClass, box,
									   NULL);

	cpshell = XtVaCreatePopupShell("cpshell", transientShellWidgetClass,
								   topLevel, NULL);
	cpform = XtVaCreateManagedWidget("cpform", formWidgetClass, cpshell, NULL);
	cptitle = XtVaCreateManagedWidget("cptitle", labelWidgetClass, cpform,
									  NULL);
	cplabel1 = XtVaCreateManagedWidget("cplabel1", labelWidgetClass, cpform,
									   NULL);
	cpchar = XtVaCreateManagedWidget("cpchar", toggleWidgetClass, cpform,
									 NULL);
	cpline = XtVaCreateManagedWidget("cpline", toggleWidgetClass, cpform,
									 NULL);
	cplabel2 = XtVaCreateManagedWidget("cplabel2", labelWidgetClass, cpform,
									   NULL);
	cpkdf = XtVaCreateManagedWidget("cpkdf", commandWidgetClass, cpform, NULL);
	cplabel3 = XtVaCreateManagedWidget("cplabel3", labelWidgetClass, cpform,
									   NULL);
	cpktt = XtVaCreateManagedWidget("cpktt", commandWidgetClass, cpform, NULL);
	cpokay = XtVaCreateManagedWidget("cpokay", commandWidgetClass, cpform,
									 NULL);
	cpcancel = XtVaCreateManagedWidget("cpcancel", commandWidgetClass, cpform,
									   NULL);
	cplistshell = XtVaCreatePopupShell("cplistshell",
									   overrideShellWidgetClass, cpshell,
									   NULL);
	cplist = XtVaCreateManagedWidget("cplist", listWidgetClass, cplistshell,
									 NULL);

	XtAddCallback(cpkdf, XtNcallback, (XtCallbackProc)CPPopupFileList, KDFEXT);
	XtAddCallback(cpktt, XtNcallback, (XtCallbackProc)CPPopupFileList, KTTEXT);
	XtAddCallback(cplist, XtNcallback, (XtCallbackProc)CPFileChosen, NULL);
	XtAddCallback(cpokay, XtNcallback, CPOkay, NULL);
	XtAddCallback(cpcancel, XtNcallback, CPCancel, NULL);

	XtVaGetValues(korEngStat,
				  XtNjohabFont, &jf, XtNinternalWidth, &iw,
				  XtNinternalHeight, &ih, NULL);
	XtVaSetValues(korEngStat,
				  XtNwidth, jf->max_bounds.width + 2*iw,
				  XtNheight,
				  jf->max_bounds.ascent+jf->max_bounds.descent + 2 * ih, NULL);
}

static void parse_resources()
{
	KeySym keysym;
	ModifierMaskType modifier, modifier_mask;
	
	if (app_data.transport == NULL &&
		app_data.toggle_key == NULL &&
		app_data.chinese_key == NULL &&
		app_data.popup_key == NULL)
		print_e(0, "The app-defaults file seems to be missing.\n");
	if (app_data.transport == NULL)
		print_e(1, "No transport mechanism specified.");
	if (app_data.toggle_key == NULL)
		print_e(1, "No Korean-English toggle key specified.");
	if (app_data.chinese_key == NULL)
		print_e(1, "No Chinese conversion key specified.");
	if (app_data.popup_key == NULL)
		print_e(1, "No Popup key specified.");

	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.toggle_key))
		print_e(1, XtNkorEngToggleKey " resource malformed.");
	else {
		trigger_keys[TOGGLE].keysym = keysym;
		trigger_keys[TOGGLE].modifier = modifier;
		trigger_keys[TOGGLE].modifier_mask = modifier_mask;
	}
	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.popup_key))
		print_e(1, XtNpopupKey " resource malformed.");
	else {
		trigger_keys[POPUP].keysym = keysym;
		trigger_keys[POPUP].modifier = modifier;
		trigger_keys[POPUP].modifier_mask = modifier_mask;
	}
	if (!parse_key(&keysym, &modifier, &modifier_mask, app_data.chinese_key))
		print_e(1, XtNchineseConversionKey " resource malformed.");
	else {
		trigger_keys[CONVERSION].keysym = keysym;
		trigger_keys[CONVERSION].modifier = modifier;
		trigger_keys[CONVERSION].modifier_mask = modifier_mask;
	}
}

static void open_om()
{
	XrmDatabase rdb = XrmGetDatabase(display);

	if ((om = XOpenOM(display, rdb, program_name, APPCLASS)) == NULL)
		print_e(1, "Cannot open output method.\n");
}
	
static void help_msg(char *argv[])
{
	print_e(0, "%s understands all standard Xt command-line options.\n"
			"Additional options are as follows:\n"
			"-transport <transport connection name>\n",
			program_name);
}

static int protocol_handler(XIMS ims, IMProtocol *call_data)
{
	IContext *target;
	ICIDType icid;
	
	switch (call_data->major_code) {
	case XIM_OPEN:
		return open_handler(ims, &call_data->imopen);
	case XIM_CLOSE:
		return close_handler(ims, &call_data->imclose);
	case XIM_CREATE_IC:
		return create_ic_handler(ims, &call_data->changeic);
	case XIM_DESTROY_IC:
		return destroy_ic_handler(ims, &call_data->destroyic);
	default:
		// Each struct of the following protocols have icid as its
		// fourth field.  So icid will have the correct value whether
		// the request is XIM_RESET_IC or not.
		icid = call_data->resetic.icid;
		
		if ((target = [theICManager findICWithID:icid]))
			switch (call_data->major_code) {
			case XIM_SET_IC_FOCUS:
				return [target setICFocus:&call_data->changefocus];
			case XIM_UNSET_IC_FOCUS:
				return [target unsetICFocus:&call_data->changefocus];
			case XIM_RESET_IC:
				return [target resetIC:&call_data->resetic];
			case XIM_SET_IC_VALUES:
				return [target setICValues:&call_data->changeic];
			case XIM_GET_IC_VALUES:
				return [target getICValues:&call_data->changeic];
			case XIM_TRIGGER_NOTIFY:
				return [target triggerNotify:&call_data->triggernotify];
			case XIM_FORWARD_EVENT:
				return [target forwardEvent:&call_data->forwardevent];
			case XIM_PREEDIT_START_REPLY:
				return [target preeditStartReply:&call_data->preedit_callback];
			case XIM_PREEDIT_CARET_REPLY:
				return [target preeditCaretReply:&call_data->preedit_callback];
			case XIM_STR_CONVERSION_REPLY:
				return
					[target strConversionReply:&call_data->strconv_callback];
			default:
				print_e(0, "Unknown protocol number, %d, encountered.\n",
						call_data->major_code);
			}
		else
			print_e(0, "Strange IC id.\n");
	}
	
	return False;
}

static int open_handler(XIMS ims, IMOpenStruct *cd)
{
#ifdef DEBUG
	print_e(0, "open_handler(): connect_id = %d, lang = %s.\n",
		   (int)cd->connect_id, cd->lang.name);
#endif
	return True;
}

static int close_handler(XIMS ims, IMCloseStruct *cd)
{
#ifdef DEBUG
	print_e(0, "close_handler(): connect_id = %d.\n", (int)cd->connect_id);
#endif
	return True;
}

static int create_ic_handler(XIMS ims, IMChangeICStruct *cd)
{
	IContext *anIC = [[IContext alloc] initWith:cd];

	if (anIC == nil ||
		[theICManager add:anIC] == nil)
	{
		[anIC free];
		print_e(0, "Input context initialization failed.\n");
		return False;
	}

	// IC id is returned to the client through this structure.
	cd->icid = [anIC getID];
	
	return True;
}

static int destroy_ic_handler(XIMS ims, IMDestroyICStruct *cd)
{
	if ([lastFocusedIC getID] == cd->icid)
		lastFocusedIC = nil;
	[theICManager deleteICWithID:cd->icid];

	return True;
}

//
//  utility functions
//

int trigger_key_index(XKeyEvent *kev)
{
	int i;

	for (i = 0; i < XtNumber(trigger_keys); i++)
		if (XLookupKeysym(kev, 0) == trigger_keys[i].keysym &&
			(kev->state & trigger_keys[i].modifier_mask) ==
			trigger_keys[i].modifier)
			return i;

	return -1;
}

void print_e(int exit_code, char *format, ...)
{
	va_list ap;

	if (format == NULL)
		perror(program_name);
	else {
		va_start(ap, format);
		fprintf(stderr, "%s: ", program_name);
		vfprintf(stderr, format, ap);
		va_end(ap);
	}

	if (exit_code != 0)
		exit(exit_code);
}

//
//  callback procedures
//

/* ARGSUSED */
static void CPPopupFileList(Widget w,
							char *file_ext, XtPointer call_data)
{
	char *p, *q;
	char directory[FILENAME_MAX];
	
	if (!file_list) {
		if ((file_list = malloc(3 * sizeof(file_list[0]))) == NULL)
			return;
		file_list_size = 3;
	}
	
	for (nfiles = 0, p = app_data.search_path; *p; p = (*q ? q + 1 : q)) {
		DIR *ds;
		struct dirent *e;
		
		if ((q = strchr(p, ':')) == NULL)
			q = p + strlen(p);
		memcpy(directory, p, q - p);
		directory[q - p] = '\0';
		if ((ds = opendir(directory)) == NULL) {
			print_e(0, "Cannot open directory: %s\n", directory);
			continue;
		}
		while ((e = readdir(ds)) != NULL) {
			if (has_extension(e->d_name, file_ext) &&
				access(strcat(strcat(directory, "/"), e->d_name), R_OK) == 0)
			{
				if (nfiles >= file_list_size) {
					char **new_file_list;
					int new_file_list_size = file_list_size + file_list_size/2;

					if ((new_file_list =
						 realloc(file_list, new_file_list_size *
								 sizeof(new_file_list[0]))) == NULL)
						goto sort;
					file_list = new_file_list;
					file_list_size = new_file_list_size;
				}

				if ((file_list[nfiles] = strdup(e->d_name)) == NULL)
					goto sort;
				nfiles++;
			}
			directory[q - p] = '\0';
		}
		closedir(ds);
	} // for

sort:
	if (nfiles > 0) {
		Position x, y;
		int width;
		static int compare_file_names(char **a, char **b);

		qsort(file_list, nfiles, sizeof(file_list[0]),
			  (QsortComp)compare_file_names);
		XtVaSetValues(cplist,
					  XtNlist, file_list, XtNnumberStrings, nfiles, NULL);
		XtTranslateCoords(w, 0, 0, &x, &y);
		XtVaGetValues(w, XtNwidth, &width, NULL);
		XtVaSetValues(cplistshell, XtNx, x + width, XtNy, y, NULL);
		XtPopupSpringLoaded(cplistshell);
		XtGrabPointer(cplistshell, True, ButtonReleaseMask,
					  GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
	}
}

static int compare_file_names(char **a, char **b)
{
	return strcmp(*a, *b);
}

/* ARGSUSED */
static void CPFileChosen(Widget w, XtPointer client_data,
						 XawListReturnStruct *call_data)
{
	if (has_extension(call_data->string, KDFEXT))
		XtVaSetValues(cpkdf, XtNlabel, call_data->string, NULL);
	else if (has_extension(call_data->string, KTTEXT))
		XtVaSetValues(cpktt, XtNlabel, call_data->string, NULL);
}

/* ARGSUSED */
static void CPOkay(Widget w, XtPointer client_data, XtPointer call_data)
{
	[lastFocusedIC changeParameters];
	XtPopdown(cpshell);
	cp_popped_up = NO;
}

/* ARGSUSED */
static void CPCancel(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(cpshell);
	cp_popped_up = NO;
}
