#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <X11/Xlib.h>

#include "sutil.h"

#define DELIMITERS " \f\n\r\t\v"

static struct {
	char *name;
	unsigned int mask;
} modifier_table [] = {
	{ "Shift",   ShiftMask },
	{ "Control", ControlMask },
	{ "Mod1", 	 Mod1Mask },
	{ "Mod2", 	 Mod2Mask },
	{ "Mod3", 	 Mod3Mask },
	{ "Mod4", 	 Mod4Mask },
	{ "Mod5", 	 Mod5Mask }
};

char *strmemcpy(char *dest, char *src, int n)
{
	memcpy(dest, src, n);
	dest[n] = '\0';
	
	return dest;
}

int memstrcmp(unsigned char *m,  int n, unsigned char*s)
{
	for (; n > 0 && *s; m++, s++, n--)
		if (*m != *s)
			return *m - *s;

	if (*s)
		return - *s;
	else
		return n;
}

char *next_line(char *p, char *eof)
{
	if ((p = next_occurrence_of('\n', p, eof)) < eof)
		p++;

	return p;
}

char *eat_space(char *p, char *eof)
{
	for (; p < eof && isspace(*p) && *p != '\n'; p++)
		;

	return p;
}

char *next_occurrence_of(int c, char *p, char *eof)
{
	for (; p < eof && *p != c; p++)
		;
	
	return p;
}

char *end_of_word(char *p, char *eof)
{
	for (; p < eof && (isalnum(*p) || *p == '_'); p++)
		;

	return p;
}

/*
 *  returns pointer to the next character of the first nonspace
 *  to the left of (and not including) *p.
 */
char *previous_occurrence_of_nonspace(char *p, char *bof)
{
	for (; p > bof && isspace(*(p-1)); p--)
		;

	return p;
}

int parse_key(KeySym *keysym,
			  unsigned int *modifier,unsigned int *modifier_mask,
			  char *key_spec)
{
	char *p;
	char *s;
	KeySym new_keysym;
	unsigned int new_modifier = 0, new_modifier_mask = 0;

	/* strtok() modifies its argument string. */
	if ((s = strdup(key_spec)) == NULL) {
		perror("parse_key()");
		return 0;
	}

	if ((p = strtok(s, DELIMITERS)) == NULL) /* No keysym string */
		goto fail;
	if ((new_keysym = XStringToKeysym(p)) == NoSymbol)
		goto fail;
	
	while ((p = strtok(NULL, DELIMITERS)) != NULL) {
		int i;
		int mod_on = 1;
		
		if (p[0] == '~') {		/* this modifier must be up. */
			mod_on = 0;
			p++;
		}
		for (i = 0; i < sizeof(modifier_table)/sizeof(modifier_table[0]); i++)
			if (strcmp(p, modifier_table[i].name) == 0) {
				new_modifier |= modifier_table[i].mask * mod_on;
				new_modifier_mask |= modifier_table[i].mask;
				break;
			}
		if (i == sizeof(modifier_table)/sizeof(modifier_table[0]))
			/* invalid modifier */
			goto fail;
	}

	*keysym = new_keysym;
	*modifier = new_modifier;
	*modifier_mask = new_modifier_mask;
	
	free(s);
	return 1;					/* okay */
	
fail:
	free(s);
	return 0;					/* not okay */
}

int has_extension(char *name, char *ext)
{
	int l1 = strlen(name), l2 = strlen(ext);

	return l1 >= l2 && strcmp(name + l1 - l2, ext) == 0;
}
