// -*- mode: objc; tab-width: 4 -*-

// The following notice is included here because text_actions[] definition
// contains part of TextAction.c of Xaw library.

/*

Copyright (c) 1989, 1994  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.

*/

#import  <objc/objc.h>
#include "Intrinsic.h"
#include <X11/Xaw/AsciiText.h>

#import  "KText.h"

id _eReceiver;					// XEvent receiver

static void MoveForwardChar(Widget, XEvent *, String *, Cardinal *);
static void MoveBackwardChar(Widget, XEvent *, String *, Cardinal *);
static void MoveForwardWord(Widget, XEvent *, String *, Cardinal *);
static void MoveBackwardWord(Widget, XEvent *, String *, Cardinal *);
static void MoveForwardParagraph(Widget, XEvent *, String *, Cardinal *);
static void MoveBackwardParagraph(Widget, XEvent *, String *, Cardinal *);
static void MoveToLineStart(Widget, XEvent *, String *, Cardinal *);
static void MoveToLineEnd(Widget, XEvent *, String *, Cardinal *);
static void MoveNextLine(Widget, XEvent *, String *, Cardinal *);
static void MovePreviousLine(Widget, XEvent *, String *, Cardinal *);
static void MoveNextPage(Widget, XEvent *, String *, Cardinal *);
static void MovePreviousPage(Widget, XEvent *, String *, Cardinal *);
static void MoveBeginningOfFile(Widget, XEvent *, String *, Cardinal *);
static void MoveEndOfFile(Widget, XEvent *, String *, Cardinal *);
static void ScrollOneLineUp(Widget, XEvent *, String *, Cardinal *);
static void ScrollOneLineDown(Widget, XEvent *, String *, Cardinal *);
static void DeleteForwardChar(Widget, XEvent *, String *, Cardinal *);
static void DeleteBackwardChar(Widget, XEvent *, String *, Cardinal *);
static void DeleteForwardWord(Widget, XEvent *, String *, Cardinal *);
static void DeleteBackwardWord(Widget, XEvent *, String *, Cardinal *);
static void DeleteCurrentSelection(Widget, XEvent *, String *, Cardinal *);
static void KillForwardWord(Widget, XEvent *, String *, Cardinal *);
static void KillBackwardWord(Widget, XEvent *, String *, Cardinal *);
static void KillCurrentSelection(Widget, XEvent *, String *, Cardinal *);
static void KillToEndOfLine(Widget, XEvent *, String *, Cardinal *);
static void KillToEndOfParagraph(Widget, XEvent *, String *, Cardinal *);
static void InsertNewLineAndIndent(Widget, XEvent *, String *, Cardinal *);
static void InsertNewLineAndBackup(Widget, XEvent *, String *, Cardinal *);
static void InsertNewLine(Widget, XEvent *, String *, Cardinal *);
static void SelectWord(Widget, XEvent *, String *, Cardinal *);
static void SelectAll(Widget, XEvent *, String *, Cardinal *);
static void SelectStart(Widget, XEvent *, String *, Cardinal *);
static void SelectAdjust(Widget, XEvent *, String *, Cardinal *);
static void SelectEnd(Widget, XEvent *, String *, Cardinal *);
static void SelectSave(Widget, XEvent *, String *, Cardinal *);
static void ExtendStart(Widget, XEvent *, String *, Cardinal *);
static void ExtendAdjust(Widget, XEvent *, String *, Cardinal *);
static void ExtendEnd(Widget, XEvent *, String *, Cardinal *);
static void InsertSelection(Widget, XEvent *, String *, Cardinal *);
static void RedrawDisplay(Widget, XEvent *, String *, Cardinal *);
static void _XawTextInsertFile(Widget, XEvent *, String *, Cardinal *);
static void _XawTextSearch(Widget, XEvent *, String *, Cardinal *);
static void InsertChar(Widget, XKeyEvent *, String *, Cardinal *);
static void InsertString(Widget, XEvent *, String *, Cardinal *);
static void TextFocusIn(Widget, XEvent *, String *, Cardinal *);
static void TextFocusOut(Widget, XEvent *, String *, Cardinal *);
static void TextEnterWindow(Widget, XEvent *, String *, Cardinal *);
static void TextLeaveWindow(Widget, XEvent *, String *, Cardinal *);
static void DisplayCaret(Widget, XEvent *, String *, Cardinal *);
static void Multiply(Widget, XEvent *, String *, Cardinal *);
static void FormParagraph(Widget, XEvent *, String *, Cardinal *);
static void TransposeCharacters(Widget, XEvent *, String *, Cardinal *);
static void NoOp(Widget, XEvent *, String *, Cardinal *);
static void _XawTextInsertFileAction(Widget, XEvent *, String *, Cardinal *);
static void _XawTextDoSearchAction(Widget, XEvent *, String *, Cardinal *);
static void _XawTextDoReplaceAction(Widget, XEvent *, String *, Cardinal *);
static void _XawTextSetField(Widget, XEvent *, String *, Cardinal *);
static void _XawTextPopdownSearchAction(Widget, XEvent *, String *, Cardinal *);
static void Reconnect(Widget, XEvent *, String *, Cardinal *);

static XtActionsRec text_actions[] = {

    // These actions are to be bound to topLevel with translation
    // mechanism.  Events delivered to the preedit Text widgets
    // and to -[IContext forwardEvent:] method are XtDispatchToWidget()ed
    // to topLevel and translated into these actions.
    // These action functions, in turn, do appropriate works and/or call
    // the real text widget actions.
    // This procedure is needed mainly to display a Korean phantom character
    // being combined just before the insertion point.

    /* motion bindings */

    {"forward-character",         MoveForwardChar},
    {"backward-character",        MoveBackwardChar},
    {"forward-word",              MoveForwardWord},
    {"backward-word",             MoveBackwardWord},
    {"forward-paragraph",         MoveForwardParagraph},
    {"backward-paragraph",        MoveBackwardParagraph},
    {"beginning-of-line",         MoveToLineStart},
    {"end-of-line",               MoveToLineEnd},
    {"next-line",                 MoveNextLine},
    {"previous-line",             MovePreviousLine},
    {"next-page",                 MoveNextPage},
    {"previous-page",             MovePreviousPage},
    {"beginning-of-file",         MoveBeginningOfFile},
    {"end-of-file",               MoveEndOfFile},
    {"scroll-one-line-up",        ScrollOneLineUp},
    {"scroll-one-line-down",      ScrollOneLineDown},

    /* delete bindings */

    {"delete-next-character",     DeleteForwardChar},
    {"delete-previous-character", DeleteBackwardChar},
    {"delete-next-word",          DeleteForwardWord},
    {"delete-previous-word",      DeleteBackwardWord},
    {"delete-selection",          DeleteCurrentSelection},

    /* kill bindings */

    {"kill-word",                 KillForwardWord},
    {"backward-kill-word",        KillBackwardWord},
    {"kill-selection",            KillCurrentSelection},
    {"kill-to-end-of-line",       KillToEndOfLine},
    {"kill-to-end-of-paragraph",  KillToEndOfParagraph},

    /* new line stuff */

    {"newline-and-indent",        InsertNewLineAndIndent},
    {"newline-and-backup",        InsertNewLineAndBackup},
    {"newline",                   InsertNewLine},

    /* Selection stuff */

    {"select-word",               SelectWord},
    {"select-all",                SelectAll},
    {"select-start",              SelectStart},
    {"select-adjust",             SelectAdjust},
    {"select-end",                SelectEnd},
    {"select-save",               SelectSave},
    {"extend-start",              ExtendStart},
    {"extend-adjust",             ExtendAdjust},
    {"extend-end",                ExtendEnd},
    {"insert-selection",          InsertSelection},

    /* Miscellaneous */

    {"redraw-display",            RedrawDisplay},
    {"insert-file",               _XawTextInsertFile},
    {"search",                    _XawTextSearch},
    {"insert-char",               (XtActionProc)InsertChar},
    {"insert-string",             InsertString},
    {"focus-in",                  TextFocusIn},
    {"focus-out",                 TextFocusOut},
    {"enter-window",              TextEnterWindow},
    {"leave-window",              TextLeaveWindow},
    {"display-caret",             DisplayCaret},
    {"multiply",                  Multiply},
    {"form-paragraph",            FormParagraph},
    {"transpose-characters",      TransposeCharacters},
    {"no-op",                     NoOp},

    /* Action to bind special translations for text Dialogs. */

    {"InsertFileAction",          _XawTextInsertFileAction},
    {"DoSearchAction",            _XawTextDoSearchAction},
    {"DoReplaceAction",           _XawTextDoReplaceAction},
    {"SetField",                  _XawTextSetField},
    {"PopdownSearchAction",       _XawTextPopdownSearchAction},

    /* Reconnect to Input Method */
    {"reconnect-im",              Reconnect} /* Li Yuhong, Omron KK, 1991 */
};

void install_text_actions(XtAppContext ac)
{
    XtAppAddActions(ac, text_actions, XtNumber(text_actions));
}

//
//  Text action functions
//

/* ARGSUSED */
static void MoveForwardChar(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textForwardCharacter:w :ev :p :np];
}

/* ARGSUSED */
static void MoveBackwardChar(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textBackwardCharacter:w :ev :p :np];
}

/* ARGSUSED */
static void MoveForwardWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textForwardWord:w :ev :p :np];
}

/* ARGSUSED */
static void MoveBackwardWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textBackwardWord:w :ev :p :np];
}

/* ARGSUSED */
static void MoveForwardParagraph(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textForwardParagraph:w :ev :p :np];
}

/* ARGSUSED */
static void MoveBackwardParagraph(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textBackwardParagraph:w :ev :p :np];
}

/* ARGSUSED */
static void MoveToLineStart(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textBeginningOfLine:w :ev :p :np];
}

/* ARGSUSED */
static void MoveToLineEnd(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textEndOfLine:w :ev :p :np];
}

/* ARGSUSED */
static void MoveNextLine(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textNextLine:w :ev :p :np];
}

/* ARGSUSED */
static void MovePreviousLine(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textPreviousLine:w :ev :p :np];
}

/* ARGSUSED */
static void MoveNextPage(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textNextPage:w :ev :p :np];
}

/* ARGSUSED */
static void MovePreviousPage(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textPreviousPage:w :ev :p :np];
}

/* ARGSUSED */
static void MoveBeginningOfFile(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textBeginningOfFile:w :ev :p :np];
}

/* ARGSUSED */
static void MoveEndOfFile(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textEndOfFile:w :ev :p :np];
}

/* ARGSUSED */
static void ScrollOneLineUp(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textScrollOneLineUp:w :ev :p :np];
}

/* ARGSUSED */
static void ScrollOneLineDown(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textScrollOneLineDown:w :ev :p :np];
}

/* ARGSUSED */
static void DeleteForwardChar(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDeleteNextCharacter:w :ev :p :np];
}

/* ARGSUSED */
static void DeleteBackwardChar(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDeletePreviousCharacter:w :ev :p :np];
}

/* ARGSUSED */
static void DeleteForwardWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDeleteNextWord:w :ev :p :np];
}

/* ARGSUSED */
static void DeleteBackwardWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDeletePreviousWord:w :ev :p :np];
}

/* ARGSUSED */
static void DeleteCurrentSelection(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDeleteSelection:w :ev :p :np];
}

static void KillForwardWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textKillWord:w :ev :p :np];
}

static void KillBackwardWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textBackwardKillWord:w :ev :p :np];
}

static void KillCurrentSelection(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textKillSelection:w :ev :p :np];
}

static void KillToEndOfLine(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textKillToEndOfLine:w :ev :p :np];
}

static void KillToEndOfParagraph(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textKillToEndOfParagraph:w :ev :p :np];
}

static void InsertNewLineAndIndent(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textNewlineAndIndent:w :ev :p :np];
}

static void InsertNewLineAndBackup(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textNewlineAndBackup:w :ev :p :np];
}

static void InsertNewLine(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textNewline:w :ev :p :np];
}

static void SelectWord(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSelectWord:w :ev :p :np];
}

static void SelectAll(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSelectAll:w :ev :p :np];
}

static void SelectStart(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSelectStart:w :ev :p :np];
}

static void SelectAdjust(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSelectAdjust:w :ev :p :np];
}

static void SelectEnd(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSelectEnd:w :ev :p :np];
}

static void SelectSave(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSelectSave:w :ev :p :np];
}

static void ExtendStart(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textExtendStart:w :ev :p :np];
}

static void ExtendAdjust(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textExtendAdjust:w :ev :p :np];
}

static void ExtendEnd(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textExtendEnd:w :ev :p :np];
}

static void InsertSelection(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textInsertSelection:w :ev :p :np];
}

static void RedrawDisplay(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textRedrawDisplay:w :ev :p :np];
}

static void _XawTextInsertFile(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textInsertFile:w :ev :p :np];
}

static void _XawTextSearch(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSearch:w :ev :p :np];
}

static void InsertChar(Widget w, XKeyEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textInsertChar:w :ev :p :np];
}

static void InsertString(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textInsertString:w :ev :p :np];
}

static void TextFocusIn(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textFocusIn:w :ev :p :np];
}

static void TextFocusOut(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textFocusOut:w :ev :p :np];
}

static void TextEnterWindow(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textEnterWindow:w :ev :p :np];
}

static void TextLeaveWindow(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textLeaveWindow:w :ev :p :np];
}

static void DisplayCaret(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDisplayCaret:w :ev :p :np];
}

static void Multiply(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textMultiply:w :ev :p :np];
}

static void FormParagraph(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textFormParagraph:w :ev :p :np];
}

static void TransposeCharacters(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textTransposeCharacters:w :ev :p :np];
}

static void NoOp(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textNoOp:w :ev :p :np];
}

static void _XawTextInsertFileAction(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textInsertFileAction:w :ev :p :np];
}

static void _XawTextDoSearchAction(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDoSearchAction:w :ev :p :np];
}

static void _XawTextDoReplaceAction(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textDoReplaceAction:w :ev :p :np];
}

static void _XawTextSetField(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textSetField:w :ev :p :np];
}

static void _XawTextPopdownSearchAction(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textPopdownSearchAction:w :ev :p :np];
}

static void Reconnect(Widget w, XEvent *ev, String *p, Cardinal *np)
{
	[_eReceiver textReconnectIm:w :ev :p :np];
}
